# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import logging
import pytest
import functools
import sys
from azure.iot.device.common.models.x509 import X509
from azure.iot.device.provisioning.security.sk_security_client import SymmetricKeySecurityClient
from azure.iot.device.provisioning.security.x509_security_client import X509SecurityClient
from azure.iot.device.provisioning.pipeline import (
    pipeline_stages_provisioning,
    pipeline_ops_provisioning,
)
from azure.iot.device.common.pipeline import pipeline_ops_base

from tests.common.pipeline.helpers import (
    assert_callback_succeeded,
    assert_callback_failed,
    all_common_ops,
    all_common_events,
    all_except,
    StageTestBase,
)
from azure.iot.device.common.pipeline import pipeline_events_base
from tests.provisioning.pipeline.helpers import all_provisioning_ops, all_provisioning_events
from tests.common.pipeline import pipeline_stage_test

logging.basicConfig(level=logging.DEBUG)

this_module = sys.modules[__name__]


# Make it look like we're always running inside pipeline threads
@pytest.fixture(autouse=True)
def apply_fake_pipeline_thread(fake_pipeline_thread):
    pass


fake_device_id = "elder_wand"
fake_registration_id = "registered_remembrall"
fake_provisioning_host = "hogwarts.com"
fake_id_scope = "weasley_wizard_wheezes"
fake_ca_cert = "fake_ca_cert"
fake_sas_token = "horcrux_token"


pipeline_stage_test.add_base_pipeline_stage_tests(
    cls=pipeline_stages_provisioning.UseSecurityClientStage,
    module=this_module,
    all_ops=all_common_ops + all_provisioning_ops,
    handled_ops=[
        pipeline_ops_provisioning.SetSymmetricKeySecurityClientOperation,
        pipeline_ops_provisioning.SetX509SecurityClientOperation,
    ],
    all_events=all_common_events + all_provisioning_events,
    handled_events=[],
)


fake_symmetric_key = "Zm9vYmFy"
fake_x509_cert_file = "fantastic_beasts"
fake_x509_cert_key_file = "where_to_find_them"
fake_pass_phrase = "alohomora"


def create_x509_security_client():
    mock_x509 = X509(fake_x509_cert_file, fake_x509_cert_key_file, fake_pass_phrase)
    return X509SecurityClient(
        provisioning_host=fake_provisioning_host,
        registration_id=fake_registration_id,
        id_scope=fake_id_scope,
        x509=mock_x509,
    )


def create_symmetric_security_client():
    return SymmetricKeySecurityClient(
        provisioning_host=fake_provisioning_host,
        registration_id=fake_registration_id,
        id_scope=fake_id_scope,
        symmetric_key=fake_symmetric_key,
    )


different_security_ops = [
    {
        "name": "set symmetric key security",
        "current_op_class": pipeline_ops_provisioning.SetSymmetricKeySecurityClientOperation,
        "security_client_function_name": create_symmetric_security_client,
    },
    {
        "name": "set x509 security",
        "current_op_class": pipeline_ops_provisioning.SetX509SecurityClientOperation,
        "security_client_function_name": create_x509_security_client,
    },
]


@pytest.fixture
def set_security_client(mocker, params_security_ops):
    # Create new security client every time to pass into fixture to avoid re-use of old security client
    # Otherwise the exception/failure raised by one test is makes the next test fail.
    op = params_security_ops["current_op_class"](
        security_client=params_security_ops["security_client_function_name"](),
        callback=mocker.MagicMock(),
    )
    return op


@pytest.mark.parametrize(
    "params_security_ops",
    different_security_ops,
    ids=[x["current_op_class"].__name__ for x in different_security_ops],
)
@pytest.mark.describe(
    "UseSecurityClientStage run_op function with SetProvisioningClientConnectionArgsOperation operations"
)
class TestUseSymmetricKeyOrX509SecurityClientRunOpWithSetSecurityClient(StageTestBase):
    @pytest.fixture
    def stage(self):
        return pipeline_stages_provisioning.UseSecurityClientStage()

    @pytest.mark.it("runs SetProvisioningClientConnectionArgsOperation op on the next stage")
    def test_runs_set_security_client_args(self, mocker, stage, set_security_client):
        stage.next._execute_op = mocker.Mock()
        stage.run_op(set_security_client)
        assert stage.next._execute_op.call_count == 1
        set_args = stage.next._execute_op.call_args[0][0]
        assert isinstance(
            set_args, pipeline_ops_provisioning.SetProvisioningClientConnectionArgsOperation
        )

    @pytest.mark.it(
        "Calls the SetSecurityClient callback with the SetProvisioningClientConnectionArgsOperation error"
        "when the SetProvisioningClientConnectionArgsOperation op raises an Exception"
    )
    def test_set_security_client_raises_exception(
        self, mocker, stage, arbitrary_exception, set_security_client
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_exception)
        stage.run_op(set_security_client)
        assert_callback_failed(op=set_security_client, error=arbitrary_exception)

    @pytest.mark.it(
        "Allows any BaseExceptions raised by SetProvisioningClientConnectionArgsOperation operations to propagate"
    )
    def test_set_security_client_raises_base_exception(
        self, mocker, stage, arbitrary_base_exception, set_security_client
    ):
        stage.next._execute_op = mocker.Mock(side_effect=arbitrary_base_exception)
        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(set_security_client)
        assert e_info.value is arbitrary_base_exception

    @pytest.mark.it(
        "Retrieves sas_token or x509_client_cert on the security_client and passes the result as the attribute of the next operation"
    )
    def test_calls_get_current_sas_token_or_get_x509_certificate(
        self, mocker, stage, set_security_client, params_security_ops
    ):
        if (
            params_security_ops["current_op_class"].__name__
            == "SetSymmetricKeySecurityClientOperation"
        ):
            spy_method = mocker.spy(set_security_client.security_client, "get_current_sas_token")
        elif params_security_ops["current_op_class"].__name__ == "SetX509SecurityClientOperation":
            spy_method = mocker.spy(set_security_client.security_client, "get_x509_certificate")

        stage.run_op(set_security_client)
        assert spy_method.call_count == 1

        set_connection_args_op = stage.next._execute_op.call_args[0][0]

        if (
            params_security_ops["current_op_class"].__name__
            == "SetSymmetricKeySecurityClientOperation"
        ):
            assert "SharedAccessSignature" in set_connection_args_op.sas_token
            assert "skn=registration" in set_connection_args_op.sas_token
            assert fake_id_scope in set_connection_args_op.sas_token
            assert fake_registration_id in set_connection_args_op.sas_token

        elif params_security_ops["current_op_class"].__name__ == "SetX509SecurityClientOperation":
            assert set_connection_args_op.client_cert.certificate_file == fake_x509_cert_file
            assert set_connection_args_op.client_cert.key_file == fake_x509_cert_key_file
            assert set_connection_args_op.client_cert.pass_phrase == fake_pass_phrase

    @pytest.mark.it(
        "Calls the callback of setting security client with no error when the next operation of "
        "etting token or setting client_cert operation succeeds"
    )
    def test_returns_success_if_set_sas_token_or_set_client_client_cert_succeeds(
        self, stage, set_security_client, next_stage_succeeds
    ):
        stage.run_op(set_security_client)
        assert_callback_succeeded(op=set_security_client)

    @pytest.mark.it(
        "Returns error when get_current_sas_token or get_x509_certificate raises an exception"
    )
    def test_get_current_sas_token_or_get_x509_certificate_raises_exception(
        self, mocker, arbitrary_exception, stage, set_security_client, params_security_ops
    ):
        if (
            params_security_ops["current_op_class"].__name__
            == "SetSymmetricKeySecurityClientOperation"
        ):
            set_security_client.security_client.get_current_sas_token = mocker.Mock(
                side_effect=arbitrary_exception
            )
        elif params_security_ops["current_op_class"].__name__ == "SetX509SecurityClientOperation":
            set_security_client.security_client.get_x509_certificate = mocker.Mock(
                side_effect=arbitrary_exception
            )
        stage.run_op(set_security_client)
        assert_callback_failed(op=set_security_client, error=arbitrary_exception)

    @pytest.mark.it(
        "Allows any BaseExceptions raised by get_current_sas_token or get_x509_certificate to propagate"
    )
    def test_get_current_sas_token_get_x509_certificate_raises_base_exception(
        self, mocker, arbitrary_base_exception, stage, set_security_client, params_security_ops
    ):
        if (
            params_security_ops["current_op_class"].__name__
            == "SetSymmetricKeySecurityClientOperation"
        ):
            set_security_client.security_client.get_current_sas_token = mocker.Mock(
                side_effect=arbitrary_base_exception
            )
        elif params_security_ops["current_op_class"].__name__ == "SetX509SecurityClientOperation":
            set_security_client.security_client.get_x509_certificate = mocker.Mock(
                side_effect=arbitrary_base_exception
            )
        with pytest.raises(arbitrary_base_exception.__class__) as e_info:
            stage.run_op(set_security_client)
        assert e_info.value is arbitrary_base_exception



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import pytest
import logging
from mock import MagicMock, patch
from threading import Timer
from azure.iot.device.iothub.auth.base_renewable_token_authentication_provider import (
    BaseRenewableTokenAuthenticationProvider,
    DEFAULT_TOKEN_VALIDITY_PERIOD,
    DEFAULT_TOKEN_RENEWAL_MARGIN,
)

logging.basicConfig(level=logging.DEBUG)


fake_signature = "__FAKE_SIGNATURE__"
fake_hostname = "__FAKE_HOSTNAME__"
fake_device_id = "__FAKE_DEVICE_ID__"
fake_module_id = "__FAKE_MODULE_ID__"
fake_current_time = 123456
fake_device_resource_uri = "{}%2Fdevices%2F{}".format(fake_hostname, fake_device_id)
fake_module_resource_uri = "{}%2Fdevices%2F{}%2Fmodules%2F{}".format(
    fake_hostname, fake_device_id, fake_module_id
)
fake_device_token_base = "SharedAccessSignature sr={}&sig={}&se=".format(
    fake_device_resource_uri, fake_signature
)
fake_module_token_base = "SharedAccessSignature sr={}&sig={}&se=".format(
    fake_module_resource_uri, fake_signature
)
new_token_validity_period = 8675
new_token_renewal_margin = 309


class FakeAuthProvider(BaseRenewableTokenAuthenticationProvider):
    def __init__(self, hostname, device_id, module_id):
        BaseRenewableTokenAuthenticationProvider.__init__(self, hostname, device_id, module_id)
        self._sign = MagicMock(return_value=fake_signature)

    def _sign(self, quoted_resource_uri, expiry):
        pass

    def parse(source):
        pass


@pytest.fixture(scope="function")
def device_auth_provider():
    return FakeAuthProvider(fake_hostname, fake_device_id, None)


@pytest.fixture(scope="function")
def module_auth_provider():
    return FakeAuthProvider(fake_hostname, fake_device_id, fake_module_id)


@pytest.fixture(scope="function")
def fake_get_current_time_function():
    with patch(
        "azure.iot.device.iothub.auth.base_renewable_token_authentication_provider.time.time",
        MagicMock(return_value=fake_current_time),
    ):
        yield


@pytest.fixture(scope="function")
def fake_timer_object():
    with patch(
        "azure.iot.device.iothub.auth.base_renewable_token_authentication_provider.Timer",
        MagicMock(spec=Timer),
    ) as PatchedTimer:
        yield PatchedTimer


def test_device_get_current_sas_token_generates_and_returns_new_sas_token(
    device_auth_provider, fake_get_current_time_function
):
    token = device_auth_provider.get_current_sas_token()
    assert device_auth_provider._sign.call_count == 1
    assert token == fake_device_token_base + str(fake_current_time + DEFAULT_TOKEN_VALIDITY_PERIOD)


def test_module_get_current_sas_token_generates_and_returns_new_sas_token(
    module_auth_provider, fake_get_current_time_function
):
    token = module_auth_provider.get_current_sas_token()
    assert module_auth_provider._sign.call_count == 1
    assert token == fake_module_token_base + str(fake_current_time + DEFAULT_TOKEN_VALIDITY_PERIOD)


def test_get_current_sas_token_returns_existing_sas_token(device_auth_provider):
    token1 = device_auth_provider.get_current_sas_token()
    token2 = device_auth_provider.get_current_sas_token()
    assert device_auth_provider._sign.call_count == 1
    assert token1 == token2


def test_generate_new_sas_token_calls_on_sas_token_updated_handler_when_sas_udpates(
    device_auth_provider
):
    update_callback = MagicMock()
    device_auth_provider.on_sas_token_updated_handler = update_callback
    device_auth_provider.generate_new_sas_token()
    update_callback.assert_called_once_with()


def test_device_generate_new_sas_token_calls_sign_with_correct_default_args(
    device_auth_provider, fake_get_current_time_function
):
    device_auth_provider.generate_new_sas_token()
    resource_uri = device_auth_provider._sign.call_args[0][0]
    expiry = device_auth_provider._sign.call_args[0][1]
    assert resource_uri == fake_device_resource_uri
    assert expiry == fake_current_time + DEFAULT_TOKEN_VALIDITY_PERIOD


def test_module_generate_new_sas_token_calls_sign_with_correct_default_args(
    module_auth_provider, fake_get_current_time_function
):
    module_auth_provider.generate_new_sas_token()
    resource_uri = module_auth_provider._sign.call_args[0][0]
    expiry = module_auth_provider._sign.call_args[0][1]
    assert resource_uri == fake_module_resource_uri
    assert expiry == fake_current_time + DEFAULT_TOKEN_VALIDITY_PERIOD


def test_generate_new_sas_token_calls_sign_with_correct_modified_expiry(
    device_auth_provider, fake_get_current_time_function
):
    device_auth_provider.token_validity_period = new_token_validity_period
    device_auth_provider.token_renewal_margin = new_token_renewal_margin
    device_auth_provider.generate_new_sas_token()
    expiry = device_auth_provider._sign.call_args[0][1]
    assert expiry == fake_current_time + new_token_validity_period


def test_generate_new_sas_token_schedules_update_timer_with_correct_default_timeout(
    device_auth_provider, fake_timer_object
):
    device_auth_provider.generate_new_sas_token()
    assert (
        fake_timer_object.call_args[0][0]
        == DEFAULT_TOKEN_VALIDITY_PERIOD - DEFAULT_TOKEN_RENEWAL_MARGIN
    )


def test_generate_new_sas_token_cancels_and_reschedules_update_timer_with_correct_modified_timeout(
    device_auth_provider, fake_timer_object
):
    device_auth_provider.token_validity_period = new_token_validity_period
    device_auth_provider.token_renewal_margin = new_token_renewal_margin
    device_auth_provider.generate_new_sas_token()
    assert fake_timer_object.call_args[0][0] == new_token_validity_period - new_token_renewal_margin


def test_update_timer_generates_new_sas_token_and_calls_on_sas_token_updated_handler(
    device_auth_provider, fake_timer_object
):
    update_callback = MagicMock()
    device_auth_provider.generate_new_sas_token()
    device_auth_provider.on_sas_token_updated_handler = update_callback
    timer_callback = fake_timer_object.call_args[0][1]
    device_auth_provider._sign.reset_mock()
    timer_callback()
    update_callback.assert_called_once_with()
    assert device_auth_provider._sign.call_count == 1


def test_finalizer_cancels_update_timer(fake_timer_object):
    # can't use the device_auth_provider fixture here because the fixture adds
    # to the object refcount and prevents del from calling the finalizer
    device_auth_provider = FakeAuthProvider(fake_hostname, fake_device_id, None)
    device_auth_provider.generate_new_sas_token()
    del device_auth_provider
    fake_timer_object.return_value.cancel.assert_called_once_with()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from azure.iot.device.iothub.pipeline import pipeline_events_iothub, pipeline_ops_iothub

all_iothub_ops = [
    pipeline_ops_iothub.SetAuthProviderOperation,
    pipeline_ops_iothub.SetIoTHubConnectionArgsOperation,
    pipeline_ops_iothub.SendD2CMessageOperation,
    pipeline_ops_iothub.SendOutputEventOperation,
]


all_iothub_events = [
    pipeline_events_iothub.C2DMessageEvent,
    pipeline_events_iothub.InputMessageEvent,
    pipeline_events_iothub.MethodRequestEvent,
]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import logging
from time import sleep
from azure.iot.device.common.evented_callback import EventedCallback

logging.basicConfig(level=logging.INFO)


@pytest.mark.describe("EventedCallback")
class TestEventedCallback(object):
    @pytest.mark.it("Can be instantiated with no args")
    def test_instantiates_without_return_arg_name(self):
        callback = EventedCallback()
        assert isinstance(callback, EventedCallback)

    @pytest.mark.it("Can be instantiated with a return_arg_name")
    def test_instantiates_with_return_arg_name(self):
        callback = EventedCallback(return_arg_name="arg_name")
        assert isinstance(callback, EventedCallback)

    @pytest.mark.it("Raises a TypeError if return_arg_name is not a string")
    def test_value_error_on_bad_return_arg_name(self):
        with pytest.raises(TypeError):
            EventedCallback(return_arg_name=1)

    @pytest.mark.it(
        "Sets the instance completion Event when a call is invoked on the instance (without return_arg_name)"
    )
    def test_calling_object_sets_event(self):
        callback = EventedCallback()
        assert not callback.completion_event.isSet()
        callback()
        sleep(0.1)  # wait to give time to complete the callback
        assert callback.completion_event.isSet()
        assert not callback.exception
        callback.wait_for_completion()

    @pytest.mark.it(
        "Sets the instance completion Event when a call is invoked on the instance (with return_arg_name)"
    )
    def test_calling_object_sets_event_with_return_arg_name(self, fake_return_arg_value):
        callback = EventedCallback(return_arg_name="arg_name")
        assert not callback.completion_event.isSet()
        callback(arg_name=fake_return_arg_value)
        sleep(0.1)  # wait to give time to complete the callback
        assert callback.completion_event.isSet()
        assert not callback.exception
        assert callback.wait_for_completion() == fake_return_arg_value

    @pytest.mark.it(
        "Raises a TypeError when a call is invoked on the instance without the correct return argument (with return_arg_name)"
    )
    def test_calling_object_raises_exception_if_return_arg_is_missing(self, fake_return_arg_value):
        callback = EventedCallback(return_arg_name="arg_name")
        with pytest.raises(TypeError):
            callback()

    @pytest.mark.it(
        "Causes an error to be raised from the wait call when an error parameter is passed to the call (without return_arg_name)"
    )
    def test_raises_error_without_return_arg_name(self, arbitrary_exception):
        callback = EventedCallback()
        assert not callback.completion_event.isSet()
        callback(error=arbitrary_exception)
        sleep(0.1)  # wait to give time to complete the callback
        assert callback.completion_event.isSet()
        assert callback.exception == arbitrary_exception
        with pytest.raises(arbitrary_exception.__class__) as e_info:
            callback.wait_for_completion()
        assert e_info.value is arbitrary_exception

    @pytest.mark.it(
        "Causes an error to be raised from the wait call when an error parameter is passed to the call (with return_arg_name)"
    )
    def test_raises_error_with_return_arg_name(self, arbitrary_exception):
        callback = EventedCallback(return_arg_name="arg_name")
        assert not callback.completion_event.isSet()
        callback(error=arbitrary_exception)
        sleep(0.1)  # wait to give time to complete the callback
        assert callback.completion_event.isSet()
        assert callback.exception == arbitrary_exception
        with pytest.raises(arbitrary_exception.__class__) as e_info:
            callback.wait_for_completion()
        assert e_info.value is arbitrary_exception



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import time
import base64
import hmac
import hashlib
import copy
import logging
import six.moves.urllib as urllib
from azure.iot.device.common.sastoken import SasToken, SasTokenError

logging.basicConfig(level=logging.DEBUG)

uri = "my.host.name"
key = "Zm9vYmFy"
key_name = "mykeyname"
device_token_kwargs = {"uri": uri, "key": key}
service_token_kwargs = {"uri": uri, "key": key, "key_name": key_name}


def generate_signature(uri, key, expiry_time):
    message = (uri + "\n" + str(expiry_time)).encode("utf-8")
    signing_key = base64.b64decode(key.encode("utf-8"))
    signed_hmac = hmac.HMAC(signing_key, message, hashlib.sha256)
    signature = urllib.parse.quote(base64.b64encode(signed_hmac.digest()))
    return signature


@pytest.fixture(params=["Device Token", "Service Token"])
def sastoken(request):
    token_type = request.param
    if token_type == "Device Token":
        return SasToken(uri, key)
    elif token_type == "Service Token":
        return SasToken(uri, key, key_name)


@pytest.mark.describe("SasToken")
class TestSasToken(object):
    @pytest.mark.it("Instantiates with default TTL of 3600 seconds")
    @pytest.mark.parametrize(
        "kwargs",
        [
            pytest.param(device_token_kwargs, id="Device Token"),
            pytest.param(service_token_kwargs, id="Service Token"),
        ],
    )
    def test_instantiates_with_default_ttl_3600(self, kwargs):
        s = SasToken(**kwargs)
        assert s._uri == kwargs.get("uri")
        assert s._key == kwargs.get("key")
        assert s._key_name == kwargs.get("key_name")
        assert s.ttl == 3600

    @pytest.mark.it("Instantiates with a custom TTL")
    @pytest.mark.parametrize(
        "kwargs",
        [
            pytest.param(device_token_kwargs, id="Device Token"),
            pytest.param(service_token_kwargs, id="Service Token"),
        ],
    )
    def test_instantiates_with_custom_ttl(self, kwargs):
        kwargs = copy.copy(kwargs)
        kwargs["ttl"] = 9000
        s = SasToken(**kwargs)
        assert s._uri == kwargs.get("uri")
        assert s._key == kwargs.get("key")
        assert s._key_name == kwargs.get("key_name")
        assert s.ttl == 9000

    @pytest.mark.it("URL encodes UTF-8 characters in provided URI")
    def test_url_encodes_utf8_characters_in_uri(self):
        utf8_uri = "my châteu.host.name"
        s = SasToken(utf8_uri, key)

        expected_uri = "my+ch%C3%A2teu.host.name"
        assert s._uri == expected_uri

    @pytest.mark.it("Raises SasTokenError if provided a key that is not base64 encoded")
    def test_raises_sastoken_error_if_key_is_not_base64(self):
        non_b64_key = "this is not base64"
        with pytest.raises(SasTokenError):
            SasToken(uri, non_b64_key)

    @pytest.mark.it("Converting object to string returns the SasToken string")
    @pytest.mark.parametrize(
        "sastoken,token_pattern",
        [
            pytest.param(
                "Device Token", "SharedAccessSignature sr={}&sig={}&se={}", id="Device Token"
            ),
            pytest.param(
                "Service Token",
                "SharedAccessSignature sr={}&sig={}&se={}&skn={}",
                id="Service Token",
            ),
        ],
        indirect=["sastoken"],
    )
    def test_string_conversion_returns_expected_sastoken_string(self, sastoken, token_pattern):
        signature = generate_signature(sastoken._uri, sastoken._key, sastoken.expiry_time)
        if sastoken._key_name:
            expected_string = token_pattern.format(
                sastoken._uri, signature, sastoken.expiry_time, sastoken._key_name
            )
        else:
            expected_string = token_pattern.format(sastoken._uri, signature, sastoken.expiry_time)
        strrep = str(sastoken)
        assert strrep == expected_string

    @pytest.mark.it("Can be refreshed to extend the expiry time by the TTL")
    def test_refreshing_token_sets_expiry_time_to_be_ttl_seconds_in_the_future(
        self, mocker, sastoken
    ):
        current_time = 1000
        mocker.patch.object(time, "time", return_value=current_time)
        sastoken.refresh()
        assert sastoken.expiry_time == current_time + sastoken.ttl

    @pytest.mark.it("Updates SasToken string upon refresh")
    def test_refreshing_token_changes_string_representation(self, sastoken):
        # This should happen because refreshing updates expiry time
        old_token_string = str(sastoken)
        time.sleep(1)
        sastoken.refresh()
        new_token_string = str(sastoken)
        assert old_token_string != new_token_string



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


class RegistrationQueryStatusResult(object):
    """
    The result of any registration attempt
    :ivar:request_id: The request id to which the response is being obtained
    :ivar:operation_id: The id of the operation as returned by the registration request.
    :ivar status: The status of the registration process as returned by provisioning service.
    Values can be "unassigned", "assigning", "assigned", "failed", "disabled"
    from the provisioning service.
    """

    def __init__(self, request_id=None, retry_after=None, operation_id=None, status=None):
        """
        :param request_id: The request id to which the response is being obtained
        :param retry_after : Number of secs after which to retry again.
        :param operation_id: The id of the operation as returned by the initial registration request.
        :param status: The status of the registration process.
        Values can be "unassigned", "assigning", "assigned", "failed", "disabled"
        from the provisioning service.
        """
        self._request_id = request_id
        self._operation_id = operation_id
        self._status = status
        self._retry_after = retry_after

    @property
    def request_id(self):
        return self._request_id

    @property
    def retry_after(self):
        return self._retry_after

    @retry_after.setter
    def retry_after(self, val):
        self._retry_after = val

    @property
    def operation_id(self):
        return self._operation_id

    @operation_id.setter
    def operation_id(self, val):
        self._operation_id = val

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, val):
        self._status = val



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import sys
from azure.iot.device.common.evented_callback import EventedCallback
from azure.iot.device.common.pipeline import (
    pipeline_stages_base,
    pipeline_ops_base,
    pipeline_stages_mqtt,
)
from . import (
    constant,
    pipeline_stages_iothub,
    pipeline_events_iothub,
    pipeline_ops_iothub,
    pipeline_stages_iothub_mqtt,
)
from azure.iot.device.iothub.auth.x509_authentication_provider import X509AuthenticationProvider

logger = logging.getLogger(__name__)


class IoTHubPipeline(object):
    def __init__(self, auth_provider, pipeline_configuration):
        """
        Constructor for instantiating a pipeline adapter object
        :param auth_provider: The authentication provider
        :param protocol: The protocol for connecting to IoT Hub
        """

        self.feature_enabled = {
            constant.C2D_MSG: False,
            constant.INPUT_MSG: False,
            constant.METHODS: False,
            constant.TWIN: False,
            constant.TWIN_PATCHES: False,
        }

        # Event Handlers - Will be set by Client after instantiation of this object
        self.on_connected = None
        self.on_disconnected = None
        self.on_c2d_message_received = None
        self.on_input_message_received = None
        self.on_method_request_received = None
        self.on_twin_patch_received = None

        # Currently a single timeout stage and a single retry stage for MQTT retry only.
        # Later, a higher level timeout and a higher level retry stage.
        self._pipeline = (
            pipeline_stages_base.PipelineRootStage(pipeline_configuration=pipeline_configuration)
            .append_stage(pipeline_stages_iothub.UseAuthProviderStage())
            .append_stage(pipeline_stages_iothub.TwinRequestResponseStage())
            .append_stage(pipeline_stages_base.CoordinateRequestAndResponseStage())
            .append_stage(pipeline_stages_iothub_mqtt.IoTHubMQTTTranslationStage())
            .append_stage(pipeline_stages_base.ReconnectStage())
            .append_stage(pipeline_stages_base.AutoConnectStage())
            .append_stage(pipeline_stages_base.ConnectionLockStage())
            .append_stage(pipeline_stages_base.RetryStage())
            .append_stage(pipeline_stages_base.OpTimeoutStage())
            .append_stage(pipeline_stages_mqtt.MQTTTransportStage())
        )

        def _on_pipeline_event(event):
            if isinstance(event, pipeline_events_iothub.C2DMessageEvent):
                if self.on_c2d_message_received:
                    self.on_c2d_message_received(event.message)
                else:
                    logger.warning("C2D message event received with no handler.  dropping.")

            elif isinstance(event, pipeline_events_iothub.InputMessageEvent):
                if self.on_input_message_received:
                    self.on_input_message_received(event.input_name, event.message)
                else:
                    logger.warning("input message event received with no handler.  dropping.")

            elif isinstance(event, pipeline_events_iothub.MethodRequestEvent):
                if self.on_method_request_received:
                    self.on_method_request_received(event.method_request)
                else:
                    logger.warning("Method request event received with no handler. Dropping.")

            elif isinstance(event, pipeline_events_iothub.TwinDesiredPropertiesPatchEvent):
                if self.on_twin_patch_received:
                    self.on_twin_patch_received(event.patch)
                else:
                    logger.warning("Twin patch event received with no handler. Dropping.")

            else:
                logger.warning("Dropping unknown pipeline event {}".format(event.name))

        def _on_connected():
            if self.on_connected:
                self.on_connected()

        def _on_disconnected():
            if self.on_disconnected:
                self.on_disconnected()

        self._pipeline.on_pipeline_event_handler = _on_pipeline_event
        self._pipeline.on_connected_handler = _on_connected
        self._pipeline.on_disconnected_handler = _on_disconnected

        callback = EventedCallback()

        if isinstance(auth_provider, X509AuthenticationProvider):
            op = pipeline_ops_iothub.SetX509AuthProviderOperation(
                auth_provider=auth_provider, callback=callback
            )
        else:  # Currently everything else goes via this block.
            op = pipeline_ops_iothub.SetAuthProviderOperation(
                auth_provider=auth_provider, callback=callback
            )

        self._pipeline.run_op(op)
        callback.wait_for_completion()

    def connect(self, callback):
        """
        Connect to the service.

        :param callback: callback which is called when the connection to the service is complete.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        logger.debug("Starting ConnectOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(pipeline_ops_base.ConnectOperation(callback=on_complete))

    def disconnect(self, callback):
        """
        Disconnect from the service.

        :param callback: callback which is called when the connection to the service has been disconnected

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        logger.debug("Starting DisconnectOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(pipeline_ops_base.DisconnectOperation(callback=on_complete))

    def send_message(self, message, callback):
        """
        Send a telemetry message to the service.

        :param message: message to send.
        :param callback: callback which is called when the message publish has been acknowledged by the service.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.SendD2CMessageOperation(message=message, callback=on_complete)
        )

    def send_output_event(self, message, callback):
        """
        Send an output message to the service.

        :param message: message to send.
        :param callback: callback which is called when the message publish has been acknowledged by the service.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.SendOutputEventOperation(message=message, callback=on_complete)
        )

    def send_method_response(self, method_response, callback):
        """
        Send a method response to the service.

        :param method_response: the method response to send
        :param callback: callback which is called when response has been acknowledged by the service

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        logger.debug("IoTHubPipeline send_method_response called")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.SendMethodResponseOperation(
                method_response=method_response, callback=on_complete
            )
        )

    def get_twin(self, callback):
        """
        Send a request for a full twin to the service.

        :param callback: callback which is called when request has been acknowledged by the service.
        This callback should have two parameters.  On success, this callback is called with the
        requested twin and error=None.  On failure, this callback is called with None for the requested
        twin and error set to the cause of the failure.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """

        def on_complete(op, error):
            if error:
                callback(error=error, twin=None)
            else:
                callback(twin=op.twin)

        self._pipeline.run_op(pipeline_ops_iothub.GetTwinOperation(callback=on_complete))

    def patch_twin_reported_properties(self, patch, callback):
        """
        Send a patch for a twin's reported properties to the service.

        :param patch: the reported properties patch to send
        :param callback: callback which is called when request has been acknowledged by the service.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.PatchTwinReportedPropertiesOperation(
                patch=patch, callback=on_complete
            )
        )

    def enable_feature(self, feature_name, callback):
        """
        Enable the given feature by subscribing to the appropriate topics.

        :param feature_name: one of the feature name constants from constant.py
        :param callback: callback which is called when the feature is enabled

        :raises: ValueError if feature_name is invalid
        """
        logger.debug("enable_feature {} called".format(feature_name))
        if feature_name not in self.feature_enabled:
            raise ValueError("Invalid feature_name")
        self.feature_enabled[feature_name] = True

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_base.EnableFeatureOperation(
                feature_name=feature_name, callback=on_complete
            )
        )

    def disable_feature(self, feature_name, callback):
        """
        Disable the given feature by subscribing to the appropriate topics.
        :param callback: callback which is called when the feature is disabled

        :param feature_name: one of the feature name constants from constant.py

        :raises: ValueError if feature_name is invalid
        """
        logger.debug("disable_feature {} called".format(feature_name))
        if feature_name not in self.feature_enabled:
            raise ValueError("Invalid feature_name")
        self.feature_enabled[feature_name] = False

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_base.DisableFeatureOperation(
                feature_name=feature_name, callback=on_complete
            )
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import sys
import pytest
import logging
from azure.iot.device.common.pipeline import pipeline_ops_base
from tests.common.pipeline import pipeline_data_object_test

this_module = sys.modules[__name__]
logging.basicConfig(level=logging.DEBUG)


@pytest.mark.describe("PipelineOperation")
class TestPipelineOperation(object):
    @pytest.mark.it("Can't be instantiated")
    def test_instantiate(self, mocker):
        with pytest.raises(TypeError):
            pipeline_ops_base.PipelineOperation(mocker.MagicMock())


pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_base.ConnectOperation, module=this_module
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_base.DisconnectOperation, module=this_module
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_base.ReconnectOperation, module=this_module
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_base.EnableFeatureOperation,
    module=this_module,
    positional_arguments=["feature_name", "callback"],
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_base.DisableFeatureOperation,
    module=this_module,
    positional_arguments=["feature_name", "callback"],
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_base.UpdateSasTokenOperation,
    module=this_module,
    positional_arguments=["sas_token", "callback"],
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_base.RequestAndResponseOperation,
    module=this_module,
    positional_arguments=[
        "request_type",
        "method",
        "resource_location",
        "request_body",
        "callback",
    ],
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_base.RequestOperation,
    module=this_module,
    positional_arguments=[
        "request_type",
        "method",
        "resource_location",
        "request_body",
        "request_id",
        "callback",
    ],
)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import pytest
import inspect

fake_count = 0


def get_next_fake_value():
    """
    return a new "unique" fake value that can be used to test that attributes
    are set correctly.  Even if we expect a particular attribute to hold something
    besides a string (like an array or an object), we can still test using a string,
    so we do.
    """
    global fake_count
    fake_count = fake_count + 1
    return "__fake_value_{}__".format(fake_count)


base_operation_defaults = {"needs_connection": False}
base_event_defaults = {}


def add_operation_test(
    cls, module, extra_defaults={}, positional_arguments=["callback"], keyword_arguments={}
):
    """
    Add a test class to test the given PipelineOperation class.  The class that
    we're testing is passed in the cls parameter, and the different initialization
    constants are passed with the named arguments that follow.
    """
    all_extra_defaults = extra_defaults.copy()
    all_extra_defaults.update(name=cls.__name__)

    add_instantiation_test(
        cls=cls,
        module=module,
        defaults=base_operation_defaults,
        extra_defaults=all_extra_defaults,
        positional_arguments=positional_arguments,
        keyword_arguments=keyword_arguments,
    )


def add_event_test(cls, module, extra_defaults={}, positional_arguments=[], keyword_arguments={}):
    """
    Add a test class to test the given PipelineOperation class.  The class that
    we're testing is passed in the cls parameter, and the different initialization
    constants are passed with the named arguments that follow.
    """
    all_extra_defaults = extra_defaults.copy()
    all_extra_defaults.update(name=cls.__name__)

    add_instantiation_test(
        cls=cls,
        module=module,
        defaults=base_event_defaults,
        extra_defaults=all_extra_defaults,
        positional_arguments=positional_arguments,
        keyword_arguments=keyword_arguments,
    )


def add_instantiation_test(
    cls, module, defaults, extra_defaults={}, positional_arguments=[], keyword_arguments={}
):
    """
    internal function that takes the class and attribute details and adds a test class which
    validates that the given class properly implements the given attributes.
    """

    # `defaults` contains an array of object attributes that should be set when
    # we call the initializer with all of the required positional arguments
    # and none of the optional keyword arguments.

    all_defaults = defaults.copy()
    for key in extra_defaults:
        all_defaults[key] = extra_defaults[key]
    for key in keyword_arguments:
        all_defaults[key] = keyword_arguments[key]

    # `args` contains an array of positional argument that we are passing to test that they
    # get assigned to the correct attribute.
    args = [get_next_fake_value() for i in range(len(positional_arguments))]

    # `kwargs` contains a dictionary of all keyword arguments, which includes required positional
    # arguments and optional keyword arguments.
    kwargs = {}
    for key in positional_arguments:
        kwargs[key] = get_next_fake_value()
    for key in keyword_arguments:
        kwargs[key] = get_next_fake_value()

    # LocalTestObject is a local class which tests the object that was passed in.  pytest doesn't test
    # against this local object, but it does test against it when we put it into the module namespace
    # for the module that was passed in.
    @pytest.mark.describe("{} - Instantiation".format(cls.__name__))
    class LocalTestObject(object):
        @pytest.mark.it(
            "Accepts {} positional arguments that get assigned to attributes of the same name: {}".format(
                len(positional_arguments), ", ".join(positional_arguments)
            )
            if len(positional_arguments)
            else "Accepts no positional arguments"
        )
        def test_positional_arguments(self):
            instance = cls(*args)
            for i in range(len(args)):
                assert getattr(instance, positional_arguments[i]) == args[i]

        @pytest.mark.it(
            "Accepts the following keyword arguments that get assigned to attributes of the same name: {}".format(
                ", ".join(kwargs.keys()) if len(kwargs) else "None"
            )
        )
        def test_keyword_arguments(self):
            instance = cls(**kwargs)
            for key in kwargs:
                assert getattr(instance, key) == kwargs[key]

        @pytest.mark.it(
            "Has the following default attributes: {}".format(
                ", ".join(["{}={}".format(key, repr(all_defaults[key])) for key in all_defaults])
            )
        )
        def test_defaults(self):
            instance = cls(*args)
            for key in all_defaults:
                if inspect.isclass(all_defaults[key]):
                    assert isinstance(getattr(instance, key), all_defaults[key])
                else:
                    assert getattr(instance, key) == all_defaults[key]

    # Adding this object to the namespace of the module that was passed in (using a name that starts with "Test")
    # will cause pytest to pick it up.
    setattr(module, "Test{}Instantiation".format(cls.__name__), LocalTestObject)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import requests
import json
import base64
import logging
import six.moves.urllib as urllib
from azure.iot.device.iothub.auth.iotedge_authentication_provider import (
    IoTEdgeAuthenticationProvider,
    IoTEdgeHsm,
    IoTEdgeError,
)
from .shared_auth_tests import SharedBaseRenewableAuthenticationProviderInstantiationTests
from azure.iot.device import constant

logging.basicConfig(level=logging.DEBUG)


@pytest.fixture
def gateway_hostname():
    return "__FAKE_GATEWAY_HOSTNAME__"


@pytest.fixture
def module_generation_id():
    return "__FAKE_MODULE_GENERATION_ID__"


@pytest.fixture
def workload_uri():
    return "http://__FAKE_WORKLOAD_URI__/"


@pytest.fixture
def api_version():
    return "__FAKE_API_VERSION__"


@pytest.fixture
def certificate():
    return "__FAKE_CERTIFICATE__"


@pytest.fixture
def mock_hsm(mocker, certificate):
    mock_hsm = mocker.patch(
        "azure.iot.device.iothub.auth.iotedge_authentication_provider.IoTEdgeHsm"
    ).return_value
    mock_hsm.get_trust_bundle.return_value = certificate
    return mock_hsm


@pytest.fixture
def hsm(module_id, module_generation_id, workload_uri, api_version):
    return IoTEdgeHsm(
        module_id=module_id,
        module_generation_id=module_generation_id,
        workload_uri=workload_uri,
        api_version=api_version,
    )


@pytest.fixture
def auth_provider(
    mock_hsm,
    hostname,
    device_id,
    module_id,
    gateway_hostname,
    module_generation_id,
    workload_uri,
    api_version,
):
    return IoTEdgeAuthenticationProvider(
        hostname=hostname,
        device_id=device_id,
        module_id=module_id,
        gateway_hostname=gateway_hostname,
        module_generation_id=module_generation_id,
        workload_uri=workload_uri,
        api_version=api_version,
    )


#######################################
# IoTEdgeAuthenticationProvider Tests #
#######################################


@pytest.mark.describe("IoTEdgeAuthenticationProvider - Instantiation")
class TestIoTEdgeAuthenticationProviderInstantiation(
    SharedBaseRenewableAuthenticationProviderInstantiationTests
):

    # TODO: Increase coverage by completing parent class

    @pytest.mark.it("Sets the gateway_hostname parameter as an instance attribute")
    def test_gateway_hostname(self, auth_provider, gateway_hostname):
        assert auth_provider.gateway_hostname == gateway_hostname

    @pytest.mark.it("Creates an instance of the IoTEdgeHsm")
    def test_creates_edge_hsm(self, auth_provider, mock_hsm):
        assert auth_provider.hsm is mock_hsm

    @pytest.mark.it(
        "Sets a certificate acquired from the IoTEdgeHsm as the ca_cert instance attribute"
    )
    def test_ca_cert_from_edge_hsm(self, auth_provider, mock_hsm):
        assert auth_provider.ca_cert is mock_hsm.get_trust_bundle.return_value
        assert mock_hsm.get_trust_bundle.call_count == 1


# TODO: Potentially get rid of this test class depending on how the parent class is tested/refactored.
# After all, we really shouldn't be testing convention-private methods.
@pytest.mark.describe("IoTEdgeAuthenticationProvider - ._sign()")
class TestIoTEdgeAuthenticationProviderSign(object):
    @pytest.mark.it("Requests signing of a string in the format '<quoted_resource_uri>/n<expiry>'")
    def test_sign_request(self, mocker, auth_provider, mock_hsm):
        uri = "my/resource/uri"
        expiry = 1234567
        string_to_sign = uri + "\n" + str(expiry)

        auth_provider._sign(uri, expiry)

        assert mock_hsm.sign.call_count == 1
        assert mock_hsm.sign.call_args == mocker.call(string_to_sign)

    @pytest.mark.it("Returns the signed string provided by the IoTEdgeHsm")
    def test_returns_signed_response(self, auth_provider, mock_hsm):
        uri = "my/resource/uri"
        expiry = 1234567

        signed_string = auth_provider._sign(uri, expiry)

        assert signed_string is mock_hsm.sign.return_value


####################
# IoTEdgeHsm Tests #
####################


@pytest.mark.describe("IoTEdgeHsm - Instantiation")
class TestIoTEdgeHsmInstantiation(object):
    @pytest.mark.it("URL encodes the module_id parameter and sets it as an instance attribute")
    def test_module_id(self, module_generation_id, workload_uri, api_version):
        my_module_id = "not url //encoded"
        expected_module_id = urllib.parse.quote(my_module_id)
        hsm = IoTEdgeHsm(
            module_id=my_module_id,
            module_generation_id=module_generation_id,
            workload_uri=workload_uri,
            api_version=api_version,
        )

        assert my_module_id != expected_module_id
        assert hsm.module_id == expected_module_id

    @pytest.mark.it("Sets the module_generation_id paramater as an instance attribute")
    def test_module_generation_id(self, hsm, module_generation_id):
        assert hsm.module_generation_id == module_generation_id

    @pytest.mark.it(
        "Converts the workload_uri parameter into requests-unixsocket format and sets it as an instance attribute"
    )
    def test_workload_uri(self, module_id, module_generation_id, api_version):
        my_workload_uri = "unix:///var/run/iotedge/workload.sock"
        expected_workload_uri = "http+unix://%2Fvar%2Frun%2Fiotedge%2Fworkload.sock/"
        hsm = IoTEdgeHsm(
            module_id=module_id,
            module_generation_id=module_generation_id,
            workload_uri=my_workload_uri,
            api_version=api_version,
        )

        assert hsm.workload_uri == expected_workload_uri

    @pytest.mark.it("Sets the api_version paramater as an instance attribute")
    def test_api_version(self, hsm, api_version):
        assert hsm.api_version == api_version


@pytest.mark.describe("IoTEdgeHsm - .get_trust_bundle()")
class TestIoTEdgeHsmGetTrustBundle(object):
    @pytest.mark.it("Makes an HTTP request to EdgeHub for the trust bundle")
    def test_requests_trust_bundle(self, mocker, hsm):
        mock_request_get = mocker.patch.object(requests, "get")
        expected_url = hsm.workload_uri + "trust-bundle"
        expected_params = {"api-version": hsm.api_version}
        expected_headers = {"User-Agent": urllib.parse.quote_plus(constant.USER_AGENT)}

        hsm.get_trust_bundle()

        assert mock_request_get.call_count == 1
        assert mock_request_get.call_args == mocker.call(
            expected_url, params=expected_params, headers=expected_headers
        )

    @pytest.mark.it("Returns the certificate from the trust bundle received from EdgeHub")
    def test_returns_received_trust_bundle(self, mocker, hsm, certificate):
        mock_request_get = mocker.patch.object(requests, "get")
        mock_response = mock_request_get.return_value
        mock_response.json.return_value = {"certificate": certificate}

        cert = hsm.get_trust_bundle()

        assert cert is certificate

    @pytest.mark.it("Raises IoTEdgeError if a bad request is made to EdgeHub")
    def test_bad_request(self, mocker, hsm):
        mock_request_get = mocker.patch.object(requests, "get")
        mock_response = mock_request_get.return_value
        error = requests.exceptions.HTTPError()
        mock_response.raise_for_status.side_effect = error

        with pytest.raises(IoTEdgeError) as e_info:
            hsm.get_trust_bundle()
        assert e_info.value.__cause__ is error

    @pytest.mark.it("Raises IoTEdgeError if there is an error in json decoding the trust bundle")
    def test_bad_json(self, mocker, hsm):
        mock_request_get = mocker.patch.object(requests, "get")
        mock_response = mock_request_get.return_value
        error = ValueError()
        mock_response.json.side_effect = error

        with pytest.raises(IoTEdgeError) as e_info:
            hsm.get_trust_bundle()
        assert e_info.value.__cause__ is error

    @pytest.mark.it("Raises IoTEdgeError if the certificate is missing from the trust bundle")
    def test_bad_trust_bundle(self, mocker, hsm):
        mock_request_get = mocker.patch.object(requests, "get")
        mock_response = mock_request_get.return_value
        # Return an empty json dict with no 'certificate' key
        mock_response.json.return_value = {}

        with pytest.raises(IoTEdgeError):
            hsm.get_trust_bundle()


@pytest.mark.describe("IoTEdgeHsm - .sign()")
class TestIoTEdgeHsmSign(object):
    @pytest.mark.it("Makes an HTTP request to EdgeHub to sign a piece of string data")
    def test_requests_data_signing(self, mocker, hsm):
        data_str = "somedata"
        data_str_b64 = "c29tZWRhdGE="
        mock_request_post = mocker.patch.object(requests, "post")
        mock_request_post.return_value.json.return_value = {"digest": "somedigest"}
        expected_url = "{workload_uri}modules/{module_id}/genid/{module_generation_id}/sign".format(
            workload_uri=hsm.workload_uri,
            module_id=hsm.module_id,
            module_generation_id=hsm.module_generation_id,
        )
        expected_params = {"api-version": hsm.api_version}
        expected_headers = {"User-Agent": urllib.parse.quote_plus(constant.USER_AGENT)}
        expected_json = json.dumps({"keyId": "primary", "algo": "HMACSHA256", "data": data_str_b64})

        hsm.sign(data_str)

        assert mock_request_post.call_count == 1
        assert mock_request_post.call_args == mocker.call(
            url=expected_url, params=expected_params, headers=expected_headers, data=expected_json
        )

    @pytest.mark.it("Base64 encodes the string data in the request")
    def test_b64_encodes_data(self, mocker, hsm):
        # This test is actually implicitly tested in the first test, but it's
        # important to have an explicit test for it since it's a requirement
        data_str = "somedata"
        data_str_b64 = base64.b64encode(data_str.encode("utf-8")).decode()
        mock_request_post = mocker.patch.object(requests, "post")
        mock_request_post.return_value.json.return_value = {"digest": "somedigest"}

        hsm.sign(data_str)

        sent_data = json.loads(mock_request_post.call_args[1]["data"])["data"]

        assert data_str != data_str_b64
        assert sent_data == data_str_b64

    @pytest.mark.it("Returns the signed data received from EdgeHub")
    def test_returns_signed_data(self, mocker, hsm):
        expected_digest = "somedigest"
        mock_request_post = mocker.patch.object(requests, "post")
        mock_request_post.return_value.json.return_value = {"digest": expected_digest}

        signed_data = hsm.sign("somedata")

        assert signed_data == expected_digest

    @pytest.mark.it("URL encodes the signed data before returning it")
    def test_url_encodes_signed_data(self, mocker, hsm):
        raw_signed_data = "this digest will be encoded"
        expected_signed_data = urllib.parse.quote(raw_signed_data)
        mock_request_post = mocker.patch.object(requests, "post")
        mock_request_post.return_value.json.return_value = {"digest": raw_signed_data}

        signed_data = hsm.sign("somedata")

        assert raw_signed_data != expected_signed_data
        assert signed_data == expected_signed_data

    @pytest.mark.it("Raises IoTEdgeError if a bad request is made to EdgeHub")
    def test_bad_request(self, mocker, hsm):
        mock_request_post = mocker.patch.object(requests, "post")
        mock_response = mock_request_post.return_value
        error = requests.exceptions.HTTPError()
        mock_response.raise_for_status.side_effect = error

        with pytest.raises(IoTEdgeError) as e_info:
            hsm.sign("somedata")
        assert e_info.value.__cause__ is error

    @pytest.mark.it("Raises IoTEdgeError if there is an error in json decoding the signed response")
    def test_bad_json(self, mocker, hsm):
        mock_request_post = mocker.patch.object(requests, "post")
        mock_response = mock_request_post.return_value
        error = ValueError()
        mock_response.json.side_effect = error
        with pytest.raises(IoTEdgeError) as e_info:
            hsm.sign("somedata")
        assert e_info.value.__cause__ is error

    @pytest.mark.it("Raises IoTEdgeError if the signed data is missing from the response")
    def test_bad_response(self, mocker, hsm):
        mock_request_post = mocker.patch.object(requests, "post")
        mock_response = mock_request_post.return_value
        mock_response.json.return_value = {}

        with pytest.raises(IoTEdgeError):
            hsm.sign("somedata")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import logging
from azure.iot.device.common.models import X509
from azure.iot.device.common.pipeline import pipeline_stages_base
from azure.iot.device.provisioning.security.sk_security_client import SymmetricKeySecurityClient
from azure.iot.device.provisioning.security.x509_security_client import X509SecurityClient
from azure.iot.device.provisioning.pipeline.provisioning_pipeline import ProvisioningPipeline
from azure.iot.device.provisioning.pipeline import pipeline_ops_provisioning
from tests.common.pipeline import helpers
import json

logging.basicConfig(level=logging.DEBUG)

send_msg_qos = 1

fake_symmetric_key = "Zm9vYmFy"
fake_registration_id = "MyPensieve"
fake_id_scope = "Enchanted0000Ceiling7898"
fake_provisioning_host = "beauxbatons.academy-net"
fake_device_id = "elder_wand"
fake_registration_id = "registered_remembrall"
fake_provisioning_host = "hogwarts.com"
fake_id_scope = "weasley_wizard_wheezes"
fake_ca_cert = "fake_certificate"
fake_sas_token = "horcrux_token"
fake_security_client = "secure_via_muffliato"
fake_request_id = "fake_request_1234"
fake_mqtt_payload = "hello hogwarts"
fake_register_publish_payload = '{{"payload": {json_payload}, "registrationId": "{reg_id}"}}'.format(
    reg_id=fake_registration_id, json_payload=json.dumps(fake_mqtt_payload)
)
fake_operation_id = "fake_operation_9876"
fake_sub_unsub_topic = "$dps/registrations/res/#"
fake_x509_cert_file = "fantastic_beasts"
fake_x509_cert_key_file = "where_to_find_them"
fake_pass_phrase = "alohomora"


def mock_x509():
    return X509(fake_x509_cert_file, fake_x509_cert_key_file, fake_pass_phrase)


different_security_clients = [
    pytest.param(
        {
            "client_class": SymmetricKeySecurityClient,
            "init_kwargs": {
                "provisioning_host": fake_provisioning_host,
                "registration_id": fake_registration_id,
                "id_scope": fake_id_scope,
                "symmetric_key": fake_symmetric_key,
            },
            "set_args_op_class": pipeline_ops_provisioning.SetSymmetricKeySecurityClientOperation,
        },
        id="Symmetric",
    ),
    pytest.param(
        {
            "client_class": X509SecurityClient,
            "init_kwargs": {
                "provisioning_host": fake_provisioning_host,
                "registration_id": fake_registration_id,
                "id_scope": fake_id_scope,
                "x509": mock_x509(),
            },
            "set_args_op_class": pipeline_ops_provisioning.SetX509SecurityClientOperation,
        },
        id="X509",
    ),
]


def assert_for_symmetric_key(password):
    assert password is not None
    assert "SharedAccessSignature" in password
    assert "skn=registration" in password
    assert fake_id_scope in password
    assert fake_registration_id in password


def assert_for_client_x509(x509):
    assert x509 is not None
    assert x509.certificate_file == fake_x509_cert_file
    assert x509.key_file == fake_x509_cert_key_file
    assert x509.pass_phrase == fake_pass_phrase


@pytest.fixture(scope="function")
def input_security_client(params_security_clients):
    return params_security_clients["client_class"](**params_security_clients["init_kwargs"])


@pytest.fixture
def pipeline_configuration(mocker):
    return mocker.MagicMock()


# automatically mock the transport for all tests in this file.
@pytest.fixture(autouse=True)
def mock_mqtt_transport(mocker):
    return mocker.patch(
        "azure.iot.device.provisioning.pipeline.provisioning_pipeline.pipeline_stages_mqtt.MQTTTransport"
    ).return_value


@pytest.fixture(scope="function")
def mock_provisioning_pipeline(
    mocker, input_security_client, mock_mqtt_transport, pipeline_configuration
):
    provisioning_pipeline = ProvisioningPipeline(input_security_client, pipeline_configuration)
    provisioning_pipeline.on_connected = mocker.MagicMock()
    provisioning_pipeline.on_disconnected = mocker.MagicMock()
    provisioning_pipeline.on_message_received = mocker.MagicMock()
    helpers.add_mock_method_waiter(provisioning_pipeline, "on_connected")
    helpers.add_mock_method_waiter(provisioning_pipeline, "on_disconnected")
    helpers.add_mock_method_waiter(mock_mqtt_transport, "publish")

    yield provisioning_pipeline
    provisioning_pipeline.disconnect()


@pytest.mark.parametrize("params_security_clients", different_security_clients)
@pytest.mark.describe("Provisioning pipeline - Initializer")
class TestInit(object):
    @pytest.mark.it("Happens correctly with the specific security client")
    def test_instantiates_correctly(
        self, params_security_clients, input_security_client, pipeline_configuration
    ):
        provisioning_pipeline = ProvisioningPipeline(input_security_client, pipeline_configuration)
        assert provisioning_pipeline._pipeline is not None

    @pytest.mark.it("Calls the correct op to pass the security client args into the pipeline")
    def test_passes_security_client_args(
        self, mocker, params_security_clients, input_security_client, pipeline_configuration
    ):
        mocker.spy(pipeline_stages_base.PipelineRootStage, "run_op")
        provisioning_pipeline = ProvisioningPipeline(input_security_client, pipeline_configuration)

        op = provisioning_pipeline._pipeline.run_op.call_args[0][1]
        assert provisioning_pipeline._pipeline.run_op.call_count == 1
        assert isinstance(op, params_security_clients["set_args_op_class"])
        assert op.security_client is input_security_client

    @pytest.mark.it("Raises an exception if the pipeline op to set security client args fails")
    def test_passes_security_client_args_failure(
        self,
        mocker,
        params_security_clients,
        input_security_client,
        arbitrary_exception,
        pipeline_configuration,
    ):
        old_execute_op = pipeline_stages_base.PipelineRootStage._execute_op

        def fail_set_auth_provider(self, op):
            if isinstance(op, params_security_clients["set_args_op_class"]):
                self.complete_op(op, error=arbitrary_exception)
            else:
                old_execute_op(self, op)

        mocker.patch.object(
            pipeline_stages_base.PipelineRootStage,
            "_execute_op",
            side_effect=fail_set_auth_provider,
            autospec=True,
        )

        with pytest.raises(arbitrary_exception.__class__) as e_info:
            ProvisioningPipeline(input_security_client, pipeline_configuration)
        assert e_info.value is arbitrary_exception


@pytest.mark.parametrize("params_security_clients", different_security_clients)
@pytest.mark.describe("Provisioning pipeline - Connect")
class TestConnect(object):
    @pytest.mark.it("Calls connect on transport")
    def test_connect_calls_connect_on_provider(
        self, params_security_clients, mock_provisioning_pipeline, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()

        assert mock_mqtt_transport.connect.call_count == 1

        if params_security_clients["client_class"].__name__ == "SymmetricKeySecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is not None
            assert_for_symmetric_key(mock_mqtt_transport.connect.call_args[1]["password"])
        elif params_security_clients["client_class"].__name__ == "X509SecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is None

        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()

    @pytest.mark.it("After complete calls handler with new state")
    def test_connected_state_handler_called_wth_new_state_once_provider_gets_connected(
        self, mock_provisioning_pipeline, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()

        mock_provisioning_pipeline.on_connected.assert_called_once_with("connected")

    @pytest.mark.it("Is ignored if waiting for completion of previous one")
    def test_connect_ignored_if_waiting_for_connect_complete(
        self, mock_provisioning_pipeline, params_security_clients, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()

        assert mock_mqtt_transport.connect.call_count == 1

        if params_security_clients["client_class"].__name__ == "SymmetricKeySecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is not None
            assert_for_symmetric_key(mock_mqtt_transport.connect.call_args[1]["password"])
        elif params_security_clients["client_class"].__name__ == "X509SecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is None

        mock_provisioning_pipeline.on_connected.assert_called_once_with("connected")

    @pytest.mark.it("Is ignored if waiting for completion of send")
    def test_connect_ignored_if_waiting_for_send_complete(
        self, mock_provisioning_pipeline, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()

        mock_mqtt_transport.reset_mock()
        mock_provisioning_pipeline.on_connected.reset_mock()

        mock_provisioning_pipeline.send_request(
            request_id=fake_request_id, request_payload=fake_mqtt_payload
        )
        mock_provisioning_pipeline.connect()

        mock_mqtt_transport.connect.assert_not_called()
        mock_provisioning_pipeline.wait_for_on_connected_to_not_be_called()
        mock_provisioning_pipeline.on_connected.assert_not_called()

        mock_mqtt_transport.on_mqtt_published(0)

        mock_mqtt_transport.connect.assert_not_called()
        mock_provisioning_pipeline.wait_for_on_connected_to_not_be_called()
        mock_provisioning_pipeline.on_connected.assert_not_called()


@pytest.mark.parametrize("params_security_clients", different_security_clients)
@pytest.mark.describe("Provisioning pipeline - Send Register")
class TestSendRegister(object):
    @pytest.mark.it("Request calls publish on provider")
    def test_send_register_request_calls_publish_on_provider(
        self, mock_provisioning_pipeline, params_security_clients, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()
        mock_provisioning_pipeline.send_request(
            request_id=fake_request_id, request_payload=fake_mqtt_payload
        )

        assert mock_mqtt_transport.connect.call_count == 1

        if params_security_clients["client_class"].__name__ == "SymmetricKeySecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is not None
            assert_for_symmetric_key(mock_mqtt_transport.connect.call_args[1]["password"])
        elif params_security_clients["client_class"].__name__ == "X509SecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is None

        fake_publish_topic = "$dps/registrations/PUT/iotdps-register/?$rid={}".format(
            fake_request_id
        )

        mock_mqtt_transport.wait_for_publish_to_be_called()
        assert mock_mqtt_transport.publish.call_count == 1
        assert mock_mqtt_transport.publish.call_args[1]["topic"] == fake_publish_topic
        assert mock_mqtt_transport.publish.call_args[1]["payload"] == fake_register_publish_payload

    @pytest.mark.it("Request queues and connects before calling publish on provider")
    def test_send_request_queues_and_connects_before_sending(
        self, mock_provisioning_pipeline, params_security_clients, mock_mqtt_transport
    ):
        # send an event
        mock_provisioning_pipeline.send_request(
            request_id=fake_request_id, request_payload=fake_mqtt_payload
        )

        # verify that we called connect
        assert mock_mqtt_transport.connect.call_count == 1

        if params_security_clients["client_class"].__name__ == "SymmetricKeySecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is not None
            assert_for_symmetric_key(mock_mqtt_transport.connect.call_args[1]["password"])
        elif params_security_clients["client_class"].__name__ == "X509SecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is None

        # verify that we're not connected yet and verify that we havent't published yet
        mock_provisioning_pipeline.wait_for_on_connected_to_not_be_called()
        mock_provisioning_pipeline.on_connected.assert_not_called()
        mock_mqtt_transport.wait_for_publish_to_not_be_called()
        mock_mqtt_transport.publish.assert_not_called()

        # finish the connection
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()

        # verify that our connected callback was called and verify that we published the event
        mock_provisioning_pipeline.on_connected.assert_called_once_with("connected")

        fake_publish_topic = "$dps/registrations/PUT/iotdps-register/?$rid={}".format(
            fake_request_id
        )

        mock_mqtt_transport.wait_for_publish_to_be_called()
        assert mock_mqtt_transport.publish.call_count == 1
        assert mock_mqtt_transport.publish.call_args[1]["topic"] == fake_publish_topic
        assert mock_mqtt_transport.publish.call_args[1]["payload"] == fake_register_publish_payload

    @pytest.mark.it("Request queues and waits for connect to be completed")
    def test_send_request_queues_if_waiting_for_connect_complete(
        self, mock_provisioning_pipeline, params_security_clients, mock_mqtt_transport
    ):
        # start connecting and verify that we've called into the transport
        mock_provisioning_pipeline.connect()
        assert mock_mqtt_transport.connect.call_count == 1

        if params_security_clients["client_class"].__name__ == "SymmetricKeySecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is not None
            assert_for_symmetric_key(mock_mqtt_transport.connect.call_args[1]["password"])
        elif params_security_clients["client_class"].__name__ == "X509SecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is None

        # send an event
        mock_provisioning_pipeline.send_request(
            request_id=fake_request_id, request_payload=fake_mqtt_payload
        )

        # verify that we're not connected yet and verify that we havent't published yet
        mock_provisioning_pipeline.wait_for_on_connected_to_not_be_called()
        mock_provisioning_pipeline.on_connected.assert_not_called()
        mock_mqtt_transport.wait_for_publish_to_not_be_called()
        mock_mqtt_transport.publish.assert_not_called()

        # finish the connection
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()

        # verify that our connected callback was called and verify that we published the event
        mock_provisioning_pipeline.on_connected.assert_called_once_with("connected")
        fake_publish_topic = "$dps/registrations/PUT/iotdps-register/?$rid={}".format(
            fake_request_id
        )
        mock_mqtt_transport.wait_for_publish_to_be_called()
        assert mock_mqtt_transport.publish.call_count == 1
        assert mock_mqtt_transport.publish.call_args[1]["topic"] == fake_publish_topic
        assert mock_mqtt_transport.publish.call_args[1]["payload"] == fake_register_publish_payload

    @pytest.mark.it("Request can be sent multiple times overlapping each other")
    def test_send_request_sends_overlapped_events(
        self, mock_provisioning_pipeline, mock_mqtt_transport, mocker
    ):
        fake_request_id_1 = fake_request_id
        fake_msg_1 = fake_mqtt_payload
        fake_request_id_2 = "request_4567"
        fake_msg_2 = "Petrificus Totalus"

        # connect
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()

        # send an event
        callback_1 = mocker.MagicMock()
        mock_provisioning_pipeline.send_request(
            request_id=fake_request_id_1, request_payload=fake_msg_1, callback=callback_1
        )

        fake_publish_topic = "$dps/registrations/PUT/iotdps-register/?$rid={}".format(
            fake_request_id_1
        )
        mock_mqtt_transport.wait_for_publish_to_be_called()
        assert mock_mqtt_transport.publish.call_count == 1
        assert mock_mqtt_transport.publish.call_args[1]["topic"] == fake_publish_topic
        assert mock_mqtt_transport.publish.call_args[1]["payload"] == fake_register_publish_payload

        # while we're waiting for that send to complete, send another event
        callback_2 = mocker.MagicMock()
        # provisioning_pipeline.send_message(fake_msg_2, callback_2)
        mock_provisioning_pipeline.send_request(
            request_id=fake_request_id_2, request_payload=fake_msg_2, callback=callback_2
        )

        # verify that we've called publish twice and verify that neither send_message
        # has completed (because we didn't do anything here to complete it).
        mock_mqtt_transport.wait_for_publish_to_be_called()
        assert mock_mqtt_transport.publish.call_count == 2
        callback_1.assert_not_called()
        callback_2.assert_not_called()

    @pytest.mark.it("Connects , sends request queues and then disconnects")
    def test_connect_send_disconnect(self, mock_provisioning_pipeline, mock_mqtt_transport):
        # connect
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()

        # send an event
        mock_provisioning_pipeline.send_request(
            request_id=fake_request_id, request_payload=fake_mqtt_payload
        )
        mock_mqtt_transport.on_mqtt_published(0)

        # disconnect
        mock_provisioning_pipeline.disconnect()
        mock_mqtt_transport.disconnect.assert_called_once_with()


@pytest.mark.parametrize("params_security_clients", different_security_clients)
@pytest.mark.describe("Provisioning pipeline - Send Query")
class TestSendQuery(object):
    @pytest.mark.it("Request calls publish on provider")
    def test_send_query_calls_publish_on_provider(
        self, mock_provisioning_pipeline, params_security_clients, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()
        mock_provisioning_pipeline.send_request(
            request_id=fake_request_id,
            request_payload=fake_mqtt_payload,
            operation_id=fake_operation_id,
        )

        assert mock_mqtt_transport.connect.call_count == 1

        if params_security_clients["client_class"].__name__ == "SymmetricKeySecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is not None
            assert_for_symmetric_key(mock_mqtt_transport.connect.call_args[1]["password"])
        elif params_security_clients["client_class"].__name__ == "X509SecurityClient":
            assert mock_mqtt_transport.connect.call_args[1]["password"] is None

        fake_publish_topic = "$dps/registrations/GET/iotdps-get-operationstatus/?$rid={}&operationId={}".format(
            fake_request_id, fake_operation_id
        )

        mock_mqtt_transport.wait_for_publish_to_be_called()
        assert mock_mqtt_transport.publish.call_count == 1
        assert mock_mqtt_transport.publish.call_args[1]["topic"] == fake_publish_topic
        assert mock_mqtt_transport.publish.call_args[1]["payload"] == fake_mqtt_payload


@pytest.mark.parametrize("params_security_clients", different_security_clients)
@pytest.mark.describe("Provisioning pipeline - Disconnect")
class TestDisconnect(object):
    @pytest.mark.it("Calls disconnect on provider")
    def test_disconnect_calls_disconnect_on_provider(
        self, mock_provisioning_pipeline, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()
        mock_provisioning_pipeline.disconnect()

        mock_mqtt_transport.disconnect.assert_called_once_with()

    @pytest.mark.it("Is ignored if already disconnected")
    def test_disconnect_ignored_if_already_disconnected(
        self, mock_provisioning_pipeline, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.disconnect(None)

        mock_mqtt_transport.disconnect.assert_not_called()

    @pytest.mark.it("After complete calls handler with `disconnected` state")
    def test_disconnect_calls_client_disconnect_callback(
        self, mock_provisioning_pipeline, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()

        mock_provisioning_pipeline.disconnect()
        mock_mqtt_transport.on_mqtt_disconnected_handler(None)

        mock_provisioning_pipeline.wait_for_on_disconnected_to_be_called()
        mock_provisioning_pipeline.on_disconnected.assert_called_once_with("disconnected")


@pytest.mark.parametrize("params_security_clients", different_security_clients)
@pytest.mark.describe("Provisioning pipeline - Enable")
class TestEnable(object):
    @pytest.mark.it("Calls subscribe on provider")
    def test_subscribe_calls_subscribe_on_provider(
        self, mock_provisioning_pipeline, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()
        mock_provisioning_pipeline.enable_responses()

        assert mock_mqtt_transport.subscribe.call_count == 1
        assert mock_mqtt_transport.subscribe.call_args[1]["topic"] == fake_sub_unsub_topic


@pytest.mark.parametrize("params_security_clients", different_security_clients)
@pytest.mark.describe("Provisioning pipeline - Disable")
class TestDisable(object):
    @pytest.mark.it("Calls unsubscribe on provider")
    def test_unsubscribe_calls_unsubscribe_on_provider(
        self, mock_provisioning_pipeline, mock_mqtt_transport
    ):
        mock_provisioning_pipeline.connect()
        mock_mqtt_transport.on_mqtt_connected_handler()
        mock_provisioning_pipeline.wait_for_on_connected_to_be_called()
        mock_provisioning_pipeline.disable_responses(None)

        assert mock_mqtt_transport.unsubscribe.call_count == 1
        assert mock_mqtt_transport.unsubscribe.call_args[1]["topic"] == fake_sub_unsub_topic



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains abstract classes for the various clients of the Azure IoT Hub Device SDK
"""

import six
import abc
import logging
import os
import io
from . import auth
from . import pipeline

from azure.iot.device.iothub.pipeline.config import IoTHubPipelineConfig


logger = logging.getLogger(__name__)

# A note on implementation:
# The intializer methods accept pipeline(s) instead of an auth provider in order to protect
# the client from logic related to authentication providers. This reduces edge cases, and allows
# pipeline configuration to be specifically tailored to the method of instantiation.
# For instance, .create_from_connection_string and .create_from_edge_envrionment both can use
# SymmetricKeyAuthenticationProviders to instantiate pipeline(s), but only .create_from_edge_environment
# should use it to instantiate an EdgePipeline. If the initializer accepted an auth provider, and then
# used it to create pipelines, this detail would be lost, as there would be no way to tell if a
# SymmetricKeyAuthenticationProvider was intended to be part of an Edge scenario or not.


@six.add_metaclass(abc.ABCMeta)
class AbstractIoTHubClient(object):
    """ A superclass representing a generic IoTHub client.
    This class needs to be extended for specific clients.
    """

    def __init__(self, iothub_pipeline):
        """Initializer for a generic client.

        :param iothub_pipeline: The pipeline used to connect to the IoTHub endpoint.
        :type iothub_pipeline: :class:`azure.iot.device.iothub.pipeline.IoTHubPipeline`
        """
        self._iothub_pipeline = iothub_pipeline
        self._edge_pipeline = None

    @classmethod
    def create_from_connection_string(cls, connection_string, ca_cert=None, **kwargs):
        """
        Instantiate the client from a IoTHub device or module connection string.

        :param str connection_string: The connection string for the IoTHub you wish to connect to.
        :param str ca_cert: The trusted certificate chain. Only necessary when using a
            connection string with a GatewayHostName parameter.

        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT over websockets.
        :param str product_info: Configuration Option. Default is empty string. The string contains arbitrary product info which is appended to the user agent string.

        :raises: ValueError if given an invalid connection_string.

        :returns: An instance of an IoTHub client that uses a connection string for authentication.
        """
        # TODO: Make this device/module specific and reject non-matching connection strings.
        # This will require refactoring of the auth package to use common objects (e.g. ConnectionString)
        # in order to differentiate types of connection strings.
        authentication_provider = auth.SymmetricKeyAuthenticationProvider.parse(connection_string)
        authentication_provider.ca_cert = ca_cert  # TODO: make this part of the instantiation
        pipeline_configuration = IoTHubPipelineConfig(**kwargs)
        iothub_pipeline = pipeline.IoTHubPipeline(authentication_provider, pipeline_configuration)
        return cls(iothub_pipeline)

    @classmethod
    def create_from_shared_access_signature(cls, sas_token, **kwargs):
        """
        Instantiate the client from a Shared Access Signature (SAS) token.
        This method of instantiation is not recommended for general usage.

        :param str sas_token: The string representation of a SAS token.

        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT over websockets.
        :param str product_info: Configuration Option. Default is empty string. The string contains arbitrary product info which is appended to the user agent string.

        :raises: ValueError if given an invalid sas_token

        :returns: An instance of an IoTHub client that uses a SAS token for authentication.
        """
        authentication_provider = auth.SharedAccessSignatureAuthenticationProvider.parse(sas_token)
        pipeline_configuration = IoTHubPipelineConfig(**kwargs)
        iothub_pipeline = pipeline.IoTHubPipeline(authentication_provider, pipeline_configuration)
        return cls(iothub_pipeline)

    @abc.abstractmethod
    def connect(self):
        pass

    @abc.abstractmethod
    def disconnect(self):
        pass

    @abc.abstractmethod
    def send_message(self, message):
        pass

    @abc.abstractmethod
    def receive_method_request(self, method_name=None):
        pass

    @abc.abstractmethod
    def send_method_response(self, method_request, payload, status):
        pass

    @abc.abstractmethod
    def get_twin(self):
        pass

    @abc.abstractmethod
    def patch_twin_reported_properties(self, reported_properties_patch):
        pass

    @abc.abstractmethod
    def receive_twin_desired_properties_patch(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class AbstractIoTHubDeviceClient(AbstractIoTHubClient):
    @classmethod
    def create_from_x509_certificate(cls, x509, hostname, device_id, **kwargs):
        """
        Instantiate a client which using X509 certificate authentication.

        :param str hostname: Host running the IotHub.
            Can be found in the Azure portal in the Overview tab as the string hostname.
        :param x509: The complete x509 certificate object.
            To use the certificate the enrollment object needs to contain cert
            (either the root certificate or one of the intermediate CA certificates).
            If the cert comes from a CER file, it needs to be base64 encoded.
        :type x509: :class:`azure.iot.device.X509`
        :param str device_id: The ID used to uniquely identify a device in the IoTHub

        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT over websockets.
        :param str product_info: Configuration Option. Default is empty string. The string contains arbitrary product info which is appended to the user agent string.

        :returns: An instance of an IoTHub client that uses an X509 certificate for authentication.
        """
        authentication_provider = auth.X509AuthenticationProvider(
            x509=x509, hostname=hostname, device_id=device_id
        )
        pipeline_configuration = IoTHubPipelineConfig(**kwargs)
        iothub_pipeline = pipeline.IoTHubPipeline(authentication_provider, pipeline_configuration)
        return cls(iothub_pipeline)

    @abc.abstractmethod
    def receive_message(self):
        pass


@six.add_metaclass(abc.ABCMeta)
class AbstractIoTHubModuleClient(AbstractIoTHubClient):
    def __init__(self, iothub_pipeline, edge_pipeline=None):
        """Initializer for a module client.

        :param iothub_pipeline: The pipeline used to connect to the IoTHub endpoint.
        :type iothub_pipeline: :class:`azure.iot.device.iothub.pipeline.IoTHubPipeline`
        :param edge_pipeline: The pipeline used to connect to the Edge endpoint.
        :type edge_pipeline: :class:`azure.iot.device.iothub.pipeline.EdgePipeline`
        """
        super(AbstractIoTHubModuleClient, self).__init__(iothub_pipeline)
        self._edge_pipeline = edge_pipeline

    @classmethod
    def create_from_edge_environment(cls, **kwargs):
        """
        Instantiate the client from the IoT Edge environment.

        This method can only be run from inside an IoT Edge container, or in a debugging
        environment configured for Edge development (e.g. Visual Studio, Visual Studio Code)

        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT over websockets.
        :param str product_info: Configuration Option. Default is empty string. The string contains arbitrary product info which is appended to the user agent string.

        :raises: OSError if the IoT Edge container is not configured correctly.
        :raises: ValueError if debug variables are invalid

        :returns: An instance of an IoTHub client that uses the IoT Edge environment for
            authentication.
        """
        # First try the regular Edge container variables
        try:
            hostname = os.environ["IOTEDGE_IOTHUBHOSTNAME"]
            device_id = os.environ["IOTEDGE_DEVICEID"]
            module_id = os.environ["IOTEDGE_MODULEID"]
            gateway_hostname = os.environ["IOTEDGE_GATEWAYHOSTNAME"]
            module_generation_id = os.environ["IOTEDGE_MODULEGENERATIONID"]
            workload_uri = os.environ["IOTEDGE_WORKLOADURI"]
            api_version = os.environ["IOTEDGE_APIVERSION"]
        except KeyError:
            # As a fallback, try the Edge local dev variables for debugging.
            # These variables are set by VS/VS Code in order to allow debugging
            # of Edge application code in a non-Edge dev environment.
            try:
                connection_string = os.environ["EdgeHubConnectionString"]
                ca_cert_filepath = os.environ["EdgeModuleCACertificateFile"]
            except KeyError as e:
                new_err = OSError("IoT Edge environment not configured correctly")
                new_err.__cause__ = e
                raise new_err
            # TODO: variant ca_cert file vs data object that would remove the need for this fopen
            # Read the certificate file to pass it on as a string
            try:
                with io.open(ca_cert_filepath, mode="r") as ca_cert_file:
                    ca_cert = ca_cert_file.read()
            except (OSError, IOError) as e:
                # In Python 2, a non-existent file raises IOError, and an invalid file raises an IOError.
                # In Python 3, a non-existent file raises FileNotFoundError, and an invalid file raises an OSError.
                # However, FileNotFoundError inherits from OSError, and IOError has been turned into an alias for OSError,
                # thus we can catch the errors for both versions in this block.
                # Unfortunately, we can't distinguish cause of error from error type, so the raised ValueError has a generic
                # message. If, in the future, we want to add detail, this could be accomplished by inspecting the e.errno
                # attribute
                new_err = ValueError("Invalid CA certificate file")
                new_err.__cause__ = e
                raise new_err
            # Use Symmetric Key authentication for local dev experience.
            try:
                authentication_provider = auth.SymmetricKeyAuthenticationProvider.parse(
                    connection_string
                )
            except ValueError:
                raise
            authentication_provider.ca_cert = ca_cert
        else:
            # Use an HSM for authentication in the general case
            try:
                authentication_provider = auth.IoTEdgeAuthenticationProvider(
                    hostname=hostname,
                    device_id=device_id,
                    module_id=module_id,
                    gateway_hostname=gateway_hostname,
                    module_generation_id=module_generation_id,
                    workload_uri=workload_uri,
                    api_version=api_version,
                )
            except auth.IoTEdgeError as e:
                new_err = OSError("Unexpected failure in IoTEdge")
                new_err.__cause__ = e
                raise new_err

        pipeline_configuration = IoTHubPipelineConfig(**kwargs)
        iothub_pipeline = pipeline.IoTHubPipeline(authentication_provider, pipeline_configuration)
        edge_pipeline = pipeline.EdgePipeline(authentication_provider)
        return cls(iothub_pipeline, edge_pipeline=edge_pipeline)

    @classmethod
    def create_from_x509_certificate(cls, x509, hostname, device_id, module_id, **kwargs):
        """
        Instantiate a client which using X509 certificate authentication.

        :param str hostname: Host running the IotHub.
            Can be found in the Azure portal in the Overview tab as the string hostname.
        :param x509: The complete x509 certificate object.
            To use the certificate the enrollment object needs to contain cert
            (either the root certificate or one of the intermediate CA certificates).
            If the cert comes from a CER file, it needs to be base64 encoded.
        :type x509: :class:`azure.iot.device.X509`
        :param str device_id: The ID used to uniquely identify a device in the IoTHub
        :param str module_id: The ID used to uniquely identify a module on a device on the IoTHub.

        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT over websockets.
        :param str product_info: Configuration Option. Default is empty string. The string contains arbitrary product info which is appended to the user agent string.

        :returns: An instance of an IoTHub client that uses an X509 certificate for authentication.
        """
        authentication_provider = auth.X509AuthenticationProvider(
            x509=x509, hostname=hostname, device_id=device_id, module_id=module_id
        )
        pipeline_configuration = IoTHubPipelineConfig(**kwargs)
        iothub_pipeline = pipeline.IoTHubPipeline(authentication_provider, pipeline_configuration)
        return cls(iothub_pipeline)

    @abc.abstractmethod
    def send_message_to_output(self, message, output_name):
        pass

    @abc.abstractmethod
    def receive_message_on_input(self, input_name):
        pass



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
import inspect
import asyncio
import logging
import azure.iot.device.common.async_adapter as async_adapter

logging.basicConfig(level=logging.DEBUG)
pytestmark = pytest.mark.asyncio


@pytest.fixture
def dummy_value():
    return 123


@pytest.fixture
def mock_function(mocker, dummy_value):
    mock_fn = mocker.MagicMock(return_value=dummy_value)
    mock_fn.__doc__ = "docstring"
    return mock_fn


@pytest.mark.describe("emulate_async()")
class TestEmulateAsync(object):
    @pytest.mark.it("Returns a coroutine function when given a function")
    async def test_returns_coroutine(self, mock_function):
        async_fn = async_adapter.emulate_async(mock_function)
        assert inspect.iscoroutinefunction(async_fn)

    @pytest.mark.it(
        "Returns a coroutine function that returns the result of the input function when called"
    )
    async def test_coroutine_returns_input_function_result(
        self, mocker, mock_function, dummy_value
    ):
        async_fn = async_adapter.emulate_async(mock_function)
        result = await async_fn(dummy_value)
        assert mock_function.call_count == 1
        assert mock_function.call_args == mocker.call(dummy_value)
        assert result == mock_function.return_value

    @pytest.mark.it("Copies the input function docstring to resulting coroutine function")
    async def test_coroutine_has_input_function_docstring(self, mock_function):
        async_fn = async_adapter.emulate_async(mock_function)
        assert async_fn.__doc__ == mock_function.__doc__

    @pytest.mark.it("Can be applied as a decorator")
    async def test_applied_as_decorator(self):

        # Define a function with emulate_async applied as a decorator
        @async_adapter.emulate_async
        def some_function():
            return "foo"

        # Call the function as a coroutine
        result = await some_function()
        assert result == "foo"


@pytest.mark.describe("AwaitableCallback")
class TestAwaitableCallback(object):
    @pytest.mark.it("Can be instantiated with no args")
    async def test_instantiates_without_return_arg_name(self):
        callback = async_adapter.AwaitableCallback()
        assert isinstance(callback, async_adapter.AwaitableCallback)

    @pytest.mark.it("Can be instantiated with a return_arg_name")
    async def test_instantiates_with_return_arg_name(self):
        callback = async_adapter.AwaitableCallback(return_arg_name="arg_name")
        assert isinstance(callback, async_adapter.AwaitableCallback)

    @pytest.mark.it("Raises a TypeError if return_arg_name is not a string")
    async def test_value_error_on_bad_return_arg_name(self):
        with pytest.raises(TypeError):
            async_adapter.AwaitableCallback(return_arg_name=1)

    @pytest.mark.it(
        "Completes the instance Future when a call is invoked on the instance (without return_arg_name)"
    )
    async def test_calling_object_completes_future(self):
        callback = async_adapter.AwaitableCallback()
        assert not callback.future.done()
        callback()
        await asyncio.sleep(0.1)  # wait to give time to complete the callback
        assert callback.future.done()
        assert not callback.future.exception()
        await callback.completion()

    @pytest.mark.it(
        "Completes the instance Future when a call is invoked on the instance (with return_arg_name)"
    )
    async def test_calling_object_completes_future_with_return_arg_name(
        self, fake_return_arg_value
    ):
        callback = async_adapter.AwaitableCallback(return_arg_name="arg_name")
        assert not callback.future.done()
        callback(arg_name=fake_return_arg_value)
        await asyncio.sleep(0.1)  # wait to give time to complete the callback
        assert callback.future.done()
        assert not callback.future.exception()
        assert await callback.completion() == fake_return_arg_value

    @pytest.mark.it(
        "Raises a TypeError when a call is invoked on the instance without the correct return argument (with return_arg_name)"
    )
    async def test_calling_object_raises_exception_if_return_arg_is_missing(
        self, fake_return_arg_value
    ):
        callback = async_adapter.AwaitableCallback(return_arg_name="arg_name")
        with pytest.raises(TypeError):
            callback()

    @pytest.mark.it(
        "Causes an error to be set on the instance Future when an error parameter is passed to the call (without return_arg_name)"
    )
    async def test_raises_error_without_return_arg_name(self, arbitrary_exception):
        callback = async_adapter.AwaitableCallback()
        assert not callback.future.done()
        callback(error=arbitrary_exception)
        await asyncio.sleep(0.1)  # wait to give time to complete the callback
        assert callback.future.done()
        assert callback.future.exception() == arbitrary_exception
        with pytest.raises(arbitrary_exception.__class__) as e_info:
            await callback.completion()
        assert e_info.value is arbitrary_exception

    @pytest.mark.it(
        "Causes an error to be set on the instance Future when an error parameter is passed to the call (with return_arg_name)"
    )
    async def test_raises_error_with_return_arg_name(self, arbitrary_exception):
        callback = async_adapter.AwaitableCallback(return_arg_name="arg_name")
        assert not callback.future.done()
        callback(error=arbitrary_exception)
        await asyncio.sleep(0.1)  # wait to give time to complete the callback
        assert callback.future.done()
        assert callback.future.exception() == arbitrary_exception
        with pytest.raises(arbitrary_exception.__class__) as e_info:
            await callback.completion()
        assert e_info.value is arbitrary_exception



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import sys
import logging
from azure.iot.device.provisioning.pipeline import pipeline_ops_provisioning
from tests.common.pipeline import pipeline_data_object_test

logging.basicConfig(level=logging.DEBUG)
this_module = sys.modules[__name__]

pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_provisioning.SetSymmetricKeySecurityClientOperation,
    module=this_module,
    positional_arguments=["security_client", "callback"],
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_provisioning.SetProvisioningClientConnectionArgsOperation,
    module=this_module,
    positional_arguments=["provisioning_host", "registration_id", "id_scope", "callback"],
    keyword_arguments={"client_cert": None, "sas_token": None},
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_provisioning.SendRegistrationRequestOperation,
    module=this_module,
    positional_arguments=["request_id", "request_payload", "registration_id", "callback"],
)
pipeline_data_object_test.add_operation_test(
    cls=pipeline_ops_provisioning.SendQueryRequestOperation,
    module=this_module,
    positional_arguments=["request_id", "operation_id", "request_payload", "callback"],
)



""" Azure IoT Device Library

This library provides clients and associated models for communicating with Azure IoT services
from an IoT device.
"""

from .iothub import *
from .provisioning import *
from .common import *
from . import iothub
from . import provisioning
from . import common
from . import patch
import sys

if sys.version_info > (3, 5):  # This only works for python 3.5+ at present
    # Dynamically patch the clients to add shim implementations for all the inherited methods.
    # This is necessary to generate accurate online docs.
    # It SHOULD not impact the functionality of the methods themselves in any way.
    patch.add_shims_for_inherited_methods(IoTHubDeviceClient)  # noqa: F405
    patch.add_shims_for_inherited_methods(IoTHubModuleClient)  # noqa: F405
    patch.add_shims_for_inherited_methods(ProvisioningDeviceClient)  # noqa: F405


# iothub and common subpackages are still showing up in intellisense

__all__ = iothub.__all__ + provisioning.__all__



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import abc
import six
import sys
import time
import uuid
import weakref
from six.moves import queue
from threading import Timer
from . import pipeline_events_base
from . import pipeline_ops_base, pipeline_ops_mqtt
from . import pipeline_thread
from . import pipeline_exceptions
from azure.iot.device.common import handle_exceptions, transport_exceptions
from azure.iot.device.common.callable_weak_method import CallableWeakMethod

logger = logging.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class PipelineStage(object):
    """
    Base class representing a stage in the processing pipeline.  Each stage is responsible for receiving
    PipelineOperation objects from the top, possibly processing them, and possibly passing them down.  It
    is also responsible for receiving PipelineEvent objects from the bottom, possibly processing them, and
    possibly passing them up.

    Each PipelineStage in the pipeline, is expected to act on some well-defined set of PipelineOperation
    types and/or some set of PipelineEvent types.  If any stage does not act on an operation or event, it
    should pass it to the next stage (for operations) or the previous stage (for events).  In this way, the
    pipeline implements the "chain of responsibility" design pattern (Gamma, et.al. "Design Patterns".
    Addison Wesley. 1995), with each stage being responsible for implementing some "rule" or "policy" of the
    pipeline, and each stage being ignorant of the stages that are before or after it in the pipeline.

    Each stage in the pipeline should act on the smallest set of rules possible, thus making stages small
    and easily testable.  Complex logic should be the exception and not the rule, and complex stages should
    operate on the most generic type of operation possible, thus allowing us to re-use complex logic for
    multiple cases.  The best way to do this is with "converter" stages that convert a specific operation to
    a more general one and with other converter stages that convert general operations to more specific ones.

    An example of a specific-to-generic stage is UseSkAuthProviderStage which takes a specific operation
    (use an auth provider) and converts it into something more generic (here is your device_id, etc, and use
    this SAS token when connecting).

    An example of a generic-to-specific stage is IoTHubMQTTTranslationStage which converts IoTHub operations
    (such as SendD2CMessageOperation) to MQTT operations (such as Publish).

    Each stage should also work in the broadest domain possible.  For example a generic stage (say
    "AutoConnectStage") that initiates a connection if any arbitrary operation needs a connection is more useful
    than having some MQTT-specific code that re-connects to the MQTT broker if the user calls Publish and
    there's no connection.

    One way to think about stages is to look at every "block of functionality" in your code and ask yourself
    "is this the one and only time I will need this code"?  If the answer is no, it might be worthwhile to
    implement that code in it's own stage in a very generic way.


    :ivar name: The name of the stage.  This is used primarily for logging
    :type name: str
    :ivar next: The next stage in the pipeline.  Set to None if this is the last stage in the pipeline.
    :type next: PipelineStage
    :ivar previous: The previous stage in the pipeline.  Set to None if this is the first stage in the pipeline.
    :type previous: PipelineStage
    :ivar pipeline_root: The first stage (root) of the pipeline.  This is useful if a stage wants to
      submit an operation to the pipeline starting at the root.  This type of behavior is uncommon but not
      unexpected.
    :type pipeline_root: PipelineStage
    """

    def __init__(self):
        """
        Initializer for PipelineStage objects.
        """
        self.name = self.__class__.__name__
        self.next = None
        self.previous = None
        self.pipeline_root = None

    @pipeline_thread.runs_on_pipeline_thread
    def run_op(self, op):
        """
        Run the given operation.  This is the public function that outside callers would call to run an
        operation.  Derived classes should override the private _execute_op function to implement
        stage-specific behavior.  When run_op returns, that doesn't mean that the operation has executed
        to completion.  Rather, it means that the pipeline has done something that will cause the
        operation to eventually execute to completion.  That might mean that something was sent over
        the network and some stage is waiting for a reply, or it might mean that the operation is sitting
        in a queue until something happens, or it could mean something entirely different.  The only
        thing you can assume is that the operation will _eventually_ complete successfully or fail, and the
        operation's callback will be called when that happens.

        :param PipelineOperation op: The operation to run.
        """
        logger.debug("{}({}): running".format(self.name, op.name))
        try:
            self._execute_op(op)
        except Exception as e:
            # This path is ONLY for unexpected errors. Expected errors should cause a fail completion
            # within ._execute_op()
            logger.error(msg="Unexpected error in {}._execute_op() call".format(self), exc_info=e)
            self.complete_op(op, error=e)

    @abc.abstractmethod
    def _execute_op(self, op):
        """
        Abstract method to run the actual operation.  This function is implemented in derived classes
        and performs the actual work that any operation expects.  The default behavior for this function
        should be to forward the event to the next stage using send_op_down for any
        operations that a particular stage might not operate on.

        See the description of the run_op method for more discussion on what it means to "run" an operation.

        :param PipelineOperation op: The operation to run.
        """
        pass

    @pipeline_thread.runs_on_pipeline_thread
    def handle_pipeline_event(self, event):
        """
        Handle a pipeline event that arrives from the stage below this stage.  Derived
        classes should not override this function.  Any stage-specific handling of
        PipelineEvent objects should be implemented by overriding the private
        _handle_pipeline_event function in the derived stage.

        :param PipelineEvent event: The event that is being passed back up the pipeline
        """
        try:
            self._handle_pipeline_event(event)
        except Exception as e:
            logger.error(
                msg="Unexpected error in {}._handle_pipeline_event() call".format(self), exc_info=e
            )
            handle_exceptions.handle_background_exception(e)

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        """
        Handle a pipeline event that arrives from the stage below this stage.  This
        is a function that is intended to be overridden in any stages that want to implement
        stage-specific handling of any events

        :param PipelineEvent event: The event that is being passed back up the pipeline
        """
        self.send_event_up(event)

    @pipeline_thread.runs_on_pipeline_thread
    def on_connected(self):
        """
        Called by lower layers when the protocol client connects
        """
        if self.previous:
            self.previous.on_connected()

    @pipeline_thread.runs_on_pipeline_thread
    def on_disconnected(self):
        """
        Called by lower layers when the protocol client disconnects
        """
        if self.previous:
            self.previous.on_disconnected()

    @pipeline_thread.runs_on_pipeline_thread
    def send_worker_op_down(self, worker_op, op):
        """
        Continue an operation using a new worker operation.  This means that the new operation
        will be passed down the pipeline (starting at the next stage). When that new
        operation completes, the original operation will be completed.  In this way,
        a stage can accept one type of operation and, effectively, change that operation
        into a different type of operation before passing it to the next stage.

        This is useful when a generic operation (such as "enable feature") needs to be
        converted into a more specific operation (such as "subscribe to mqtt topic").
        In that case, a stage's _execute_op function would call this function passing in
        the original "enable feature" op and the new "subscribe to mqtt topic"
        op.  This function will pass the "subscribe" down. When the "subscribe" op
        is completed, this function will cause the original op to complete.

        This function is only really useful if there is no data returned in the
        worker_op that that needs to be copied back into the original_op before
        completing it.  If data needs to be copied this way, some other method needs
        to be used.  (or a "copy data back" function needs to be added to this function
        as an optional parameter.)

        :param PipelineOperation op: Operation that is being continued using a
          different op.  This is most likely the operation that is currently being handled
          by the stage.  This operation is not actually continued, in that it is not
          actually passed down the pipeline.  Instead, the original_op operation is
          effectively paused while we wait for the worker_op operation to complete.  When
          the worker_op operation completes, the original_op operation will also be completed.
        :param PipelineOperation worker_op: Operation that is being passed down the pipeline
          to effectively continue the work represented by the original op.  This is most likely
          a different type of operation that is able to accomplish the intention of the
          original op in a way that is more specific than the original op.
        """

        logger.debug("{}({}): continuing with {} op".format(self.name, op.name, worker_op.name))

        @pipeline_thread.runs_on_pipeline_thread
        def worker_op_complete(worker_op, error):
            logger.debug(
                "{}({}): completing with result from {}".format(self.name, op.name, worker_op.name)
            )
            self.complete_op(op, error=error)

        worker_op.callback = worker_op_complete
        self.send_op_down(worker_op)

    @pipeline_thread.runs_on_pipeline_thread
    def send_op_down(self, op):
        """
        Helper function to continue a given operation by passing it to the next stage
        in the pipeline.  If there is no next stage in the pipeline, this function
        will fail the operation and call complete_op to return the failure back up the
        pipeline.

        :param PipelineOperation op: Operation which is being passed on
        """
        if not self.next:
            logger.error("{}({}): no next stage.  completing with error".format(self.name, op.name))
            error = pipeline_exceptions.PipelineError(
                "{} not handled after {} stage with no next stage".format(op.name, self.name)
            )
            self.complete_op(op, error=error)
        else:
            logger.debug("{}({}): passing to next stage.".format(self.name, op.name))
            self.next.run_op(op)

    @pipeline_thread.runs_on_pipeline_thread
    def complete_op(self, op, error=None):
        """
        Helper function to complete an operation by calling its callback function thus
        returning the result of the operation back up the pipeline.  This is perferred to
        calling the operation's callback directly as it provides several layers of protection
        (such as a try/except wrapper) which are strongly advised.
        """
        if error:
            logger.error("{}({}): completing with error {}".format(self.name, op.name, error))
        else:
            logger.debug("{}({}): completing without error".format(self.name, op.name))

        if op.completed:
            logger.error(
                "{}({}): completing op that has already been completed!".format(self.name, op.name)
            )
            e = pipeline_exceptions.PipelineError(
                "Internal pipeline error: attempting to complete an already-completed operation: {}({})".format(
                    self.name, op.name
                )
            )
            handle_exceptions.handle_background_exception(e)
        else:
            op.completed = True
            self.send_completed_op_up(op, error)

    @pipeline_thread.runs_on_pipeline_thread
    def send_completed_op_up(self, op, error=None):
        """
        Sends a previously-completed operation back up the pipeline, usually to the callback.
        This is used by complete_op and it's also called from code in the stage itself inside
        an intercepted return (via send_op_down_and_intercept_return).
        """
        if not op.completed:
            raise pipeline_exceptions.PipelineError(
                "Internal pipeline error: attempting to send an incomplete {} op up".format(op.name)
            )

        try:
            op.callback(op, error=error)
        except Exception as e:
            _, e, _ = sys.exc_info()
            logger.error(
                msg="Unhandled error calling back inside {}.complete_op() after {} complete".format(
                    self.name, op.name
                ),
                exc_info=e,
            )
            handle_exceptions.handle_background_exception(e)

    @pipeline_thread.runs_on_pipeline_thread
    def send_event_up(self, event):
        """
        Helper function to pass an event to the previous stage of the pipeline.  This is the default
        behavior of events while traveling through the pipeline. They start somewhere (maybe the
        bottom) and move up the pipeline until they're handled or until they error out.
        """
        if self.previous:
            logger.debug(
                "{}({}): pushing event up to {}".format(self.name, event.name, self.previous.name)
            )
            self.previous.handle_pipeline_event(event)
        else:
            logger.error("{}({}): Error: unhandled event".format(self.name, event.name))
            error = pipeline_exceptions.PipelineError(
                "{} unhandled at {} stage with no previous stage".format(event.name, self.name)
            )
            handle_exceptions.handle_background_exception(error)

    @pipeline_thread.runs_on_pipeline_thread
    def send_op_down_and_intercept_return(self, op, intercepted_return):
        """
        Function which sends an op down to the next stage in the pipeline and inserts an
        "intercepted_return" function in the return path of the op.  This way, a stage can
        continue processing of any op and use the intercepted_return function to  see the
        result of the op before returning it all the way to its original callback.  This is
        useful for stages that want to monitor the progress of ops, such as a OpTimeoutStage
        that needs to keep track of how long ops are running and when they complete.
        When that intercepted_return function is done with the op, it can use
        send_completed_op_up() to finish processing the op.
        """
        old_callback = op.callback

        def new_callback(op, error):
            op.callback = old_callback
            intercepted_return(op=op, error=error)

        op.callback = new_callback
        self.send_op_down(op)


class PipelineRootStage(PipelineStage):
    """
    Object representing the root of a pipeline.  This is where the functions to build
    the pipeline exist.  This is also where clients can add event handlers to receive
    events from the pipeline.

    :ivar on_pipeline_event_handler: Handler which can be set by users of the pipeline to
      receive PipelineEvent objects.  This is how users receive any "unsolicited"
      events from the pipeline (such as C2D messages).  This function is called with
      a PipelineEvent object every time any such event occurs.
    :type on_pipeline_event_handler: Function
    :ivar on_connected_handler: Handler which can be set by users of the pipeline to
      receive events every time the underlying transport connects
    :type on_connected_handler: Function
    :ivar on_disconnected_handler: Handler which can be set by users of the pipeline to
      receive events every time the underlying transport disconnects
    :type on_disconnected_handler: Function
    """

    def __init__(self, pipeline_configuration):
        super(PipelineRootStage, self).__init__()
        self.on_pipeline_event_handler = None
        self.on_connected_handler = None
        self.on_disconnected_handler = None
        self.connected = False
        self.pipeline_configuration = pipeline_configuration

    def run_op(self, op):
        op.callback = pipeline_thread.invoke_on_callback_thread_nowait(op.callback)
        pipeline_thread.invoke_on_pipeline_thread(super(PipelineRootStage, self).run_op)(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _execute_op(self, op):
        """
        run the operation.  At the root, the only thing to do is to pass the operation
        to the next stage.

        :param PipelineOperation op: Operation to run.
        """
        self.send_op_down(op)

    def append_stage(self, new_next_stage):
        """
        Add the next stage to the end of the pipeline.  This is the function that callers
        use to build the pipeline by appending stages.  This function returns the root of
        the pipeline so that calls to this function can be chained together.

        :param PipelineStage new_next_stage: Stage to add to the end of the pipeline
        :returns: The root of the pipeline.
        """
        old_tail = self
        while old_tail.next:
            old_tail = old_tail.next
        old_tail.next = new_next_stage
        new_next_stage.previous = old_tail
        new_next_stage.pipeline_root = self
        return self

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        """
        Override of the PipelineEvent handler.  Because this is the root of the pipeline,
        this function calls the on_pipeline_event_handler to pass the event to the
        caller.

        :param PipelineEvent event: Event to be handled, i.e. returned to the caller
          through the handle_pipeline_event (if provided).
        """
        if self.on_pipeline_event_handler:
            pipeline_thread.invoke_on_callback_thread_nowait(self.on_pipeline_event_handler)(event)
        else:
            logger.warning("incoming pipeline event with no handler.  dropping.")

    @pipeline_thread.runs_on_pipeline_thread
    def on_connected(self):
        logger.debug(
            "{}: on_connected.  on_connected_handler={}".format(
                self.name, self.on_connected_handler
            )
        )
        self.connected = True
        if self.on_connected_handler:
            pipeline_thread.invoke_on_callback_thread_nowait(self.on_connected_handler)()

    @pipeline_thread.runs_on_pipeline_thread
    def on_disconnected(self):
        logger.debug(
            "{}: on_disconnected.  on_disconnected_handler={}".format(
                self.name, self.on_disconnected_handler
            )
        )
        self.connected = False
        if self.on_disconnected_handler:
            pipeline_thread.invoke_on_callback_thread_nowait(self.on_disconnected_handler)()


class AutoConnectStage(PipelineStage):
    """
    This stage is responsible for ensuring that the protocol is connected when
    it needs to be connected.
    """

    @pipeline_thread.runs_on_pipeline_thread
    def _execute_op(self, op):
        # Any operation that requires a connection can trigger a connection if
        # we're not connected.
        if op.needs_connection and not self.pipeline_root.connected:
            logger.debug(
                "{}({}): Op needs connection.  Queueing this op and starting a ConnectionOperation".format(
                    self.name, op.name
                )
            )
            self._do_connect(op)

        # Finally, if this stage doesn't need to do anything else with this operation,
        # it just passes it down.
        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _do_connect(self, op):
        """
        Start connecting the transport in response to some operation
        """
        # function that gets called after we're connected.
        @pipeline_thread.runs_on_pipeline_thread
        def on_connect_op_complete(op_connect, error):
            if error:
                logger.error(
                    "{}({}): Connection failed.  Completing with failure because of connection failure: {}".format(
                        self.name, op.name, error
                    )
                )
                self.complete_op(op, error=error)
            else:
                logger.debug(
                    "{}({}): connection is complete.  Continuing with op".format(self.name, op.name)
                )
                self.send_op_down(op)

        # call down to the next stage to connect.
        logger.debug("{}({}): calling down with Connect operation".format(self.name, op.name))
        self.send_op_down(pipeline_ops_base.ConnectOperation(callback=on_connect_op_complete))


class ConnectionLockStage(PipelineStage):
    """
    This stage is responsible for serializing connect, disconnect, and reconnect ops on
    the pipeline, such that only a single one of these ops can go past this stage at a
    time.  This way, we don't have to worry about cases like "what happens if we try to
    disconnect if we're in the middle of reconnecting."  This stage will wait for the
    reconnect to complete before letting the disconnect past.
    """

    def __init__(self):
        super(ConnectionLockStage, self).__init__()
        self.queue = queue.Queue()
        self.blocked = False

    @pipeline_thread.runs_on_pipeline_thread
    def _execute_op(self, op):
        # If this stage is currently blocked (because we're waiting for a connection, etc,
        # to complete), we queue up all operations until after the connect completes.
        if self.blocked:
            logger.info(
                "{}({}): pipeline is blocked waiting for a prior connect/disconnect/reconnect to complete.  queueing.".format(
                    self.name, op.name
                )
            )
            self.queue.put_nowait(op)

        elif isinstance(op, pipeline_ops_base.ConnectOperation) and self.pipeline_root.connected:
            logger.info("{}({}): Transport is connected.  Completing.".format(self.name, op.name))
            self.complete_op(op)

        elif (
            isinstance(op, pipeline_ops_base.DisconnectOperation)
            and not self.pipeline_root.connected
        ):
            logger.info(
                "{}({}): Transport is disconnected.  Completing.".format(self.name, op.name)
            )
            self.complete_op(op=op)

        elif (
            isinstance(op, pipeline_ops_base.DisconnectOperation)
            or isinstance(op, pipeline_ops_base.ConnectOperation)
            or isinstance(op, pipeline_ops_base.ReconnectOperation)
        ):
            self._block(op)

            @pipeline_thread.runs_on_pipeline_thread
            def on_operation_complete(op, error):
                if error:
                    logger.error(
                        "{}({}): op failed.  Unblocking queue with error: {}".format(
                            self.name, op.name, error
                        )
                    )
                else:
                    logger.debug(
                        "{}({}): op succeeded.  Unblocking queue".format(self.name, op.name)
                    )

                self._unblock(op, error)
                logger.debug(
                    "{}({}): unblock is complete.  completing op that caused unblock".format(
                        self.name, op.name
                    )
                )
                self.send_completed_op_up(op, error)

            self.send_op_down_and_intercept_return(op, intercepted_return=on_operation_complete)

        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _block(self, op):
        """
        block this stage while we're waiting for the connect/disconnect/reconnect operation to complete.
        """
        logger.debug("{}({}): blocking".format(self.name, op.name))
        self.blocked = True

    @pipeline_thread.runs_on_pipeline_thread
    def _unblock(self, op, error):
        """
        Unblock this stage after the connect/disconnect/reconnect operation is complete.  This also means
        releasing all the operations that were queued up.
        """
        logger.debug("{}({}): unblocking and releasing queued ops.".format(self.name, op.name))
        self.blocked = False
        logger.info(
            "{}({}): processing {} items in queue".format(self.name, op.name, self.queue.qsize())
        )
        # Loop through our queue and release all the blocked operations
        # Put a new Queue in self.queue because releasing ops might put them back in the
        # queue, especially if there's a ConnectOperation in the list of ops to release
        old_queue = self.queue
        self.queue = queue.Queue()
        while not old_queue.empty():
            op_to_release = old_queue.get_nowait()
            if error:
                # if we're unblocking the queue because something (like a connect operation) failed,
                # then we fail all of the blocked operations with the same error.
                logger.error(
                    "{}({}): failing {} op because of error".format(
                        self.name, op.name, op_to_release.name
                    )
                )
                self.complete_op(op_to_release, error=error)
            else:
                logger.debug(
                    "{}({}): releasing {} op.".format(self.name, op.name, op_to_release.name)
                )
                # call run_op directly here so operations go through this stage again (especiall connect/disconnect ops)
                self.run_op(op_to_release)


class CoordinateRequestAndResponseStage(PipelineStage):
    """
    Pipeline stage which is responsible for coordinating RequestAndResponseOperation operations.  For each
    RequestAndResponseOperation operation, this stage passes down a RequestOperation operation and waits for
    an ResponseEvent event.  All other events are passed down unmodified.
    """

    def __init__(self):
        super(CoordinateRequestAndResponseStage, self).__init__()
        self.pending_responses = {}

    @pipeline_thread.runs_on_pipeline_thread
    def _execute_op(self, op):
        if isinstance(op, pipeline_ops_base.RequestAndResponseOperation):
            # Convert RequestAndResponseOperation operation into a RequestOperation operation
            # and send it down.  A lower level will convert the RequestOperation into an
            # actual protocol client operation.  The RequestAndResponseOperation operation will be
            # completed when the corresponding IotResponse event is received in this stage.

            request_id = str(uuid.uuid4())

            @pipeline_thread.runs_on_pipeline_thread
            def on_send_request_done(send_request_op, error):
                logger.debug(
                    "{}({}): Finished sending {} request to {} resource {}".format(
                        self.name, op.name, op.request_type, op.method, op.resource_location
                    )
                )
                if error:
                    logger.debug(
                        "{}({}): removing request {} from pending list".format(
                            self.name, op.name, request_id
                        )
                    )
                    del (self.pending_responses[request_id])
                    self.complete_op(op, error=error)
                else:
                    # request sent.  Nothing to do except wait for the response
                    pass

            logger.debug(
                "{}({}): Sending {} request to {} resource {}".format(
                    self.name, op.name, op.request_type, op.method, op.resource_location
                )
            )

            logger.debug(
                "{}({}): adding request {} to pending list".format(self.name, op.name, request_id)
            )
            self.pending_responses[request_id] = op

            new_op = pipeline_ops_base.RequestOperation(
                method=op.method,
                resource_location=op.resource_location,
                request_body=op.request_body,
                request_id=request_id,
                request_type=op.request_type,
                callback=on_send_request_done,
            )
            self.send_op_down(new_op)

        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _handle_pipeline_event(self, event):
        if isinstance(event, pipeline_events_base.ResponseEvent):
            # match ResponseEvent events to the saved dictionary of RequestAndResponseOperation
            # operations which have not received responses yet.  If the operation is found,
            # complete it.

            logger.debug(
                "{}({}): Handling event with request_id {}".format(
                    self.name, event.name, event.request_id
                )
            )
            if event.request_id in self.pending_responses:
                op = self.pending_responses[event.request_id]
                del (self.pending_responses[event.request_id])
                op.status_code = event.status_code
                op.response_body = event.response_body
                logger.debug(
                    "{}({}): Completing {} request to {} resource {} with status {}".format(
                        self.name,
                        op.name,
                        op.request_type,
                        op.method,
                        op.resource_location,
                        op.status_code,
                    )
                )
                self.complete_op(op)
            else:
                logger.warning(
                    "{}({}): request_id {} not found in pending list.  Nothing to do.  Dropping".format(
                        self.name, event.name, event.request_id
                    )
                )
        else:
            self.send_event_up(event)


class OpTimeoutStage(PipelineStage):
    """
    The purpose of the timeout stage is to add timeout errors to select operations

    The timeout_intervals attribute contains a list of operations to track along with
    their timeout values.  Right now this list is hard-coded but the operations and
    intervals will eventually become a parameter.

    For each operation that needs a timeout check, this stage will add a timer to
    the operation.  If the timer elapses, this stage will fail the operation with
    a PipelineTimeoutError.  The intention is that a higher stage will know what to
    do with that error and act accordingly (either return the error to the user or
    retry).

    This stage currently assumes that all timed out operation are just "lost".
    It does not attempt to cancel the operation, as Paho doesn't have a way to
    cancel an operation, and with QOS=1, sending a pub or sub twice is not
    catastrophic.

    Also, as a long-term plan, the operations that need to be watched for timeout
    will become an initialization parameter for this stage so that differet
    instances of this stage can watch for timeouts on different operations.
    This will be done because we want a lower-level timeout stage which can watch
    for timeouts at the MQTT level, and we want a higher-level timeout stage which
    can watch for timeouts at the iothub level.  In this way, an MQTT operation that
    times out can be retried as an MQTT operation and a higher-level IoTHub operation
    which times out can be retried as an IoTHub operation (which might necessitate
    redoing multiple MQTT operations).
    """

    def __init__(self):
        super(OpTimeoutStage, self).__init__()
        # use a fixed list and fixed intervals for now.  Later, this info will come in
        # as an init param or a retry poicy
        self.timeout_intervals = {
            pipeline_ops_mqtt.MQTTSubscribeOperation: 10,
            pipeline_ops_mqtt.MQTTUnsubscribeOperation: 10,
        }

    @pipeline_thread.runs_on_pipeline_thread
    def _execute_op(self, op):
        if type(op) in self.timeout_intervals:
            # Create a timer to watch for operation timeout on this op and attach it
            # to the op.
            self_weakref = weakref.ref(self)

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def on_timeout():
                this = self_weakref()
                logger.info("{}({}): returning timeout error".format(this.name, op.name))
                this.complete_op(
                    op,
                    pipeline_exceptions.PipelineTimeoutError(
                        "operation timed out before protocol client could respond"
                    ),
                )

            logger.debug("{}({}): Creating timer".format(self.name, op.name))
            op.timeout_timer = Timer(self.timeout_intervals[type(op)], on_timeout)
            op.timeout_timer.start()

            # Send the op down, but intercept the return of the op so we can
            # remove the timer when the op is done
            logger.debug("{}({}): Sending down".format(self.name, op.name))
            self.send_op_down_and_intercept_return(
                op=op, intercepted_return=self._on_intercepted_return
            )
        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _on_intercepted_return(self, op, error):
        # When an op comes back, delete the timer and pass it right up.
        if op.timeout_timer:
            op.timeout_timer.cancel()
            op.timeout_timer = None
        logger.debug("{}({}): Op completed".format(self.name, op.name))
        self.send_completed_op_up(op, error)


class RetryStage(PipelineStage):
    """
    The purpose of the retry stage is to watch specific operations for specific
    errors and retry the operations as appropriate.

    Unlike the OpTimeoutStage, this stage will never need to worry about cancelling
    failed operations.  When an operation is retried at this stage, it is already
    considered "failed", so no cancellation needs to be done.
    """

    def __init__(self):
        super(RetryStage, self).__init__()
        # Retry intervals are hardcoded for now. Later, they come in as an
        # init param, probably via retry policy.
        self.retry_intervals = {
            pipeline_ops_mqtt.MQTTSubscribeOperation: 20,
            pipeline_ops_mqtt.MQTTUnsubscribeOperation: 20,
            pipeline_ops_base.ConnectOperation: 20,
            pipeline_ops_mqtt.MQTTPublishOperation: 20,
        }
        self.ops_waiting_to_retry = []

    @pipeline_thread.runs_on_pipeline_thread
    def _execute_op(self, op):
        """
        Send all ops down and intercept their return to "watch for retry"
        """
        if self._should_watch_for_retry(op):
            self.send_op_down_and_intercept_return(
                op=op, intercepted_return=self._on_intercepted_return
            )
        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _should_watch_for_retry(self, op):
        """
        Return True if this op needs to be watched for retry.  This can be
        called before the op runs.
        """
        return type(op) in self.retry_intervals

    @pipeline_thread.runs_on_pipeline_thread
    def _should_retry(self, op, error):
        """
        Return True if this op needs to be retried.  This must be called after
        the op completes.
        """
        if error:
            if self._should_watch_for_retry(op):
                if type(error) in [
                    pipeline_exceptions.PipelineTimeoutError,
                    transport_exceptions.ConnectionDroppedError,
                    transport_exceptions.ConnectionFailedError,
                ]:
                    return True
        return False

    @pipeline_thread.runs_on_pipeline_thread
    def _on_intercepted_return(self, op, error):
        """
        Handler which gets called when operations are complete.  This function
        is where we check to see if a retry is necessary and set a "retry timer"
        which can be used to send the op down again.
        """
        if self._should_retry(op, error):
            self_weakref = weakref.ref(self)

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def do_retry():
                this = self_weakref()
                logger.info("{}({}): retrying".format(this.name, op.name))
                op.retry_timer = None
                op.completed = False
                this.ops_waiting_to_retry.remove(op)
                # Don't just send it down directly.  Instead, go through _execute_op so we get
                # retry functionality this time too
                this._execute_op(op)

            interval = self.retry_intervals[type(op)]
            logger.warning(
                "{}({}): Op needs retry with interval {} because of {}.  Setting timer.".format(
                    self.name, op.name, interval, error
                )
            )

            # if we don't keep track of this op, it might get collected.
            self.ops_waiting_to_retry.append(op)
            op.retry_timer = Timer(self.retry_intervals[type(op)], do_retry)
            op.retry_timer.start()

        else:
            if op.retry_timer:
                op.retry_timer = None
            self.send_completed_op_up(op, error)


transient_connect_errors = [
    pipeline_exceptions.OperationCancelled,
    pipeline_exceptions.PipelineTimeoutError,
    pipeline_exceptions.OperationError,
    transport_exceptions.ConnectionFailedError,
    transport_exceptions.ConnectionDroppedError,
]


class ReconnectStage(PipelineStage):
    def __init__(self):
        super(ReconnectStage, self).__init__()
        self.reconnect_timer = None
        self.virtually_connected = False
        # connect delay is hardcoded for now.  Later, this comes from a retry policy
        self.reconnect_delay = 10

    @pipeline_thread.runs_on_pipeline_thread
    def _execute_op(self, op):
        if isinstance(op, pipeline_ops_base.ConnectOperation):
            self.virtually_connected = True
            self.send_op_down(op)

        elif isinstance(op, pipeline_ops_base.DisconnectOperation):
            self.virtually_connected = False
            self.send_op_down(op)

        else:
            self.send_op_down(op)

    @pipeline_thread.runs_on_pipeline_thread
    def _set_reconnect_timer(self):
        """
        Set a timer to reconnect after some period of time
        """

        self._clear_reconnect_timer()

        self_weakref = weakref.ref(self)

        @pipeline_thread.invoke_on_pipeline_thread_nowait
        def on_reconnect_timer_expired():
            this = self_weakref()

            if not this.pipeline_root.connected:
                logger.info(
                    "{}: reconnect timer expired.  Sending connect op down".format(this.name)
                )

                def on_connect_complete(op, error):
                    inner_this = self_weakref()
                    if error:
                        if type(error) in transient_connect_errors:
                            logger.debug(
                                "{}: reconnect failed because {}.  Setting new timer.".format(
                                    inner_this.name, error
                                )
                            )
                            inner_this._set_reconnect_timer()
                        else:
                            logger.debug(
                                "{}: reconnect failed because {}.  Not setting new timer.".format(
                                    inner_this.name, error
                                )
                            )
                    else:
                        logger.debug("{}: reconnect successful".format(inner_this.name))

                logger.debug("{}: Sending connect operation down".format(this.name))
                this.send_op_down(pipeline_ops_base.ConnectOperation(on_connect_complete))
            else:
                logger.debug(
                    "{}: retry timer expired, but client is connected.  Doing nothing".format(
                        this.name
                    )
                )

        logger.info("{}: Setting reconnect timer".format(self.name))
        self.reconnect_timer = Timer(self.reconnect_delay, on_reconnect_timer_expired)
        self.reconnect_timer.start()

    @pipeline_thread.runs_on_pipeline_thread
    def _clear_reconnect_timer(self):
        """
        Clear any previous reconnect timer
        """
        if self.reconnect_timer:
            logger.info("{}: Clearing reconnect timer".format(self.name))
            self.reconnect_timer.cancel()
            self.reconnect_timer = None

    @pipeline_thread.runs_on_pipeline_thread
    def on_connected(self):
        """
        Called by lower layers when the protocol client connects
        """
        logger.debug("{}: on_connected".format(self.name))

        self._clear_reconnect_timer()

        if self.previous:
            self.previous.on_connected()

    @pipeline_thread.runs_on_pipeline_thread
    def on_disconnected(self):
        """
        Called by lower layers when the protocol client disconnects
        """
        logger.debug("{}: on_disconnected".format(self.name))
        if self.pipeline_root.connected:
            if self.virtually_connected:
                logger.info(
                    "{}: disconnected but virtually connected.  Triggering reconnect timer.".format(
                        self.name
                    )
                )
                self._set_reconnect_timer()
            else:
                logger.info(
                    "{}: disconnected, but not virtually connected.  Not triggering reconnect timer.".format(
                        self.name
                    )
                )

        if self.previous:
            self.previous.on_disconnected()
